"""Functionality for analyzing the structure of tensor networks, including
finding paths, loops, connected components, hierarchical groupings and more.
"""

import collections
import functools
import itertools
import math

from ..utils import oset, unique


class NetworkPatch:
    """A simple class to represent a patch of tensors and indices, storing
    both the tensor identifies (`tids`) and indices (`inds`) it contains.
    """

    __slots__ = ("_tids", "_inds", "_key")

    def __init__(self, tids, inds):
        self._tids = oset(tids)
        self._inds = oset(inds)
        self._key = None

    @classmethod
    def from_sequence(cls, it):
        tids = []
        inds = []
        for x in it:
            (tids if isinstance(x, int) else inds).append(x)
        return cls(tids, inds)

    @property
    def tids(self):
        return self._tids

    @property
    def inds(self):
        return self._inds

    def __iter__(self):
        return itertools.chain(self._tids, self._inds)

    @property
    def key(self):
        # build lazily as don't always need
        if self._key is None:
            self._key = frozenset(self)
        return self._key

    def merge(self, other):
        return NetworkPatch(
            tids=itertools.chain(self._tids, other._tids),
            inds=itertools.chain(self._inds, other._inds),
        )

    def __contains__(self, x):
        return x in self.key

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        if not isinstance(other, NetworkPatch):
            return NotImplemented
        return self.key == other.key

    def __repr__(self):
        return f"{self.__class__.__name__}({self._tids}, {self._inds})"


class NetworkPath(NetworkPatch):
    """A simple class to represent a path through a tensor network, storing
    both the tensor identifies (`tids`) and indices (`inds`) it passes through.
    """

    __slots__ = NetworkPatch.__slots__

    def __init__(self, tids, inds=()):
        self._tids = tuple(tids)
        self._inds = tuple(inds)
        if len(self._tids) != len(self._inds) + 1:
            raise ValueError("tids should be one longer than inds")
        self._key = None

    def __len__(self):
        return len(self._inds)

    def __iter__(self):
        # interleave tids and inds
        for tid, ind in zip(self._tids, self._inds):
            yield tid
            yield ind
        # always one more tid
        yield self._tids[-1]

    def extend(self, ind, tid):
        """Get a new path by extending this one with a new index and tensor id."""
        new = NetworkPath.__new__(NetworkPath)
        new._tids = self._tids + (tid,)
        new._inds = self._inds + (ind,)
        new._key = None if self._key is None else self._key | {tid, ind}
        return new


def istree(tn):
    """Check if this tensor network has a tree structure, (treating
    multibonds as a single edge).

    Examples
    --------

        >>> MPS_rand_state(10, 7).istree()
        True

        >>> MPS_rand_state(10, 7, cyclic=True).istree()
        False

    """
    tid0 = next(iter(tn.tensor_map))
    region = [(tid0, None)]
    seen = {tid0}
    while region:
        tid, ptid = region.pop()
        for ntid in tn._get_neighbor_tids(tid):
            if ntid == ptid:
                # ignore the previous tid we just came from
                continue
            if ntid in seen:
                # found a loop
                return False
            # expand the queue
            region.append((ntid, tid))
            seen.add(ntid)
    return True


def isconnected(tn):
    """Check whether this tensor network is connected, i.e. whether
    there is a path between any two tensors, (including size 1 indices).
    """
    tid0 = next(iter(tn.tensor_map))
    region = tn._get_subgraph_tids([tid0])
    return len(region) == len(tn.tensor_map)


def subgraphs(tn, virtual=False):
    """Split this tensor network into disconneceted subgraphs.

    Parameters
    ----------
    virtual : bool, optional
        Whether the tensor networks should view the original tensors or
        not - by default take copies.

    Returns
    -------
    list[TensorNetwork]
    """
    groups = []
    tids = oset(tn.tensor_map)

    # check all nodes
    while tids:
        # get a remaining node
        tid0 = tids.popright()
        queue = [tid0]
        group = oset(queue)

        while queue:
            # expand it until no neighbors
            tid = queue.pop()
            for tid_n in tn._get_neighbor_tids(tid):
                if tid_n in group:
                    continue
                else:
                    group.add(tid_n)
                    queue.append(tid_n)

        # remove current subgraph and continue
        tids -= group
        groups.append(group)

    return [tn._select_tids(group, virtual=virtual) for group in groups]


def get_tree_span(
    tn,
    tids,
    min_distance=0,
    max_distance=None,
    include=None,
    exclude=None,
    ndim_sort="max",
    distance_sort="min",
    sorter=None,
    weight_bonds=True,
    inwards=True,
):
    """Generate a tree on the tensor network graph, fanning out from the
    tensors identified by ``tids``, up to a maximum of ``max_distance``
    away. The tree can be visualized with
    :meth:`~quimb.tensor.tensor_core.TensorNetwork.draw_tree_span`.

    Parameters
    ----------
    tids : sequence of str
        The nodes that define the region to span out of.
    min_distance : int, optional
        Don't add edges to the tree until this far from the region. For
        example, ``1`` will not include the last merges from neighboring
        tensors in the region defined by ``tids``.
    max_distance : None or int, optional
        Terminate branches once they reach this far away. If ``None`` there
        is no limit,
    include : sequence of str, optional
        If specified, only ``tids`` specified here can be part of the tree.
    exclude : sequence of str, optional
        If specified, ``tids`` specified here cannot be part of the tree.
    ndim_sort : {'min', 'max', 'none'}, optional
        When expanding the tree, how to choose what nodes to expand to
        next, once connectivity to the current surface has been taken into
        account.
    distance_sort : {'min', 'max', 'none'}, optional
        When expanding the tree, how to choose what nodes to expand to
        next, once connectivity to the current surface has been taken into
        account.
    weight_bonds : bool, optional
        Whether to weight the 'connection' of a candidate tensor to expand
        out to using bond size as well as number of bonds.

    Returns
    -------
    list[(str, str, int)]
        The ordered list of merges, each given as tuple ``(tid1, tid2, d)``
        indicating merge ``tid1 -> tid2`` at distance ``d``.

    See Also
    --------
    draw_tree_span
    """
    # current tensors in the tree -> we will grow this
    region = oset(tids)

    # check if we should only allow a certain set of nodes
    if include is None:
        include = oset(tn.tensor_map)
    elif not isinstance(include, oset):
        include = oset(include)

    allowed = include - region

    # check if we should explicitly ignore some nodes
    if exclude is not None:
        if not isinstance(exclude, oset):
            exclude = oset(exclude)
        allowed -= exclude

    # possible merges of neighbors into the region
    candidates = []

    # actual merges we have performed, defining the tree
    merges = {}

    # distance to the original region
    distances = {tid: 0 for tid in region}

    # how many times (or weight) that neighbors are connected to the region
    connectivity = collections.defaultdict(lambda: 0)

    # given equal connectivity compare neighbors based on
    #      min/max distance and min/max ndim
    distance_coeff = {"min": -1, "max": 1, "none": 0}[distance_sort]
    ndim_coeff = {"min": -1, "max": 1, "none": 0}[ndim_sort]

    def _check_candidate(tid_surface, tid_neighb):
        """Check the expansion of ``tid_surface`` to ``tid_neighb``."""
        if (tid_neighb in region) or (tid_neighb not in allowed):
            # we've already absorbed it, or we're not allowed to
            return

        if tid_neighb not in distances:
            # defines a new spanning tree edge
            merges[tid_neighb] = tid_surface
            # graph distance to original region
            new_d = distances[tid_surface] + 1
            distances[tid_neighb] = new_d
            if (max_distance is None) or (new_d <= max_distance):
                candidates.append(tid_neighb)

        # keep track of how connected to the current surface potential new
        # nodes are
        if weight_bonds:
            connectivity[tid_neighb] += math.log2(
                tn.tensor_map[tid_surface].bonds_size(
                    tn.tensor_map[tid_neighb]
                )
            )
        else:
            connectivity[tid_neighb] += 1

    if sorter is None:

        def _sorter(t):
            # how to pick which tensor to absorb into the expanding surface
            # here, choose the candidate that is most connected to current
            # surface, breaking ties with how close it is to the original
            # tree, and how many dimensions it has
            return (
                connectivity[t],
                ndim_coeff * tn.tensor_map[t].ndim,
                distance_coeff * distances[t],
            )
    else:
        _sorter = functools.partial(
            sorter, tn=tn, distances=distances, connectivity=connectivity
        )

    # setup the initial region and candidate nodes to expand to
    for tid_surface in region:
        for tid_next in tn._get_neighbor_tids(tid_surface):
            _check_candidate(tid_surface, tid_next)

    # generate the sequence of tensor merges
    seq = []
    while candidates:
        # choose the *highest* scoring candidate
        candidates.sort(key=_sorter)
        tid_surface = candidates.pop()
        region.add(tid_surface)

        if distances[tid_surface] > min_distance:
            # checking distance allows the innermost merges to be ignored,
            # for example, to contract an environment around a region
            seq.append(
                (tid_surface, merges[tid_surface], distances[tid_surface])
            )

        # check all the neighbors of the tensor we've just expanded to
        for tid_next in tn._get_neighbor_tids(tid_surface):
            _check_candidate(tid_surface, tid_next)

    if inwards:
        # make the sequence of merges flow inwards
        seq.reverse()

    return seq


def get_local_patch(
    tn,
    tids,
    max_distance,
    include=None,
    exclude=None,
):
    """Get the local patch of tids that is within ``max_distance`` of
    the given ``tids``. This is like an unordered version of ``get_tree_span``.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to get the local patch from.
    tids : sequence of int
        The tensor ids to start from.
    max_distance : int
        The maximum distance from ``tids`` to include, in terms of graph
        distance. 0 corresponds to the original ``tids``, 1 to nearest
        neighbors and so on.
    include : sequence of int, optional
        If specified, only tids from this set can be included in the patch.
    exclude : sequence of int, optional
        If specified, tids from this set cannot be included in the patch.

    Returns
    -------
    tuple[int]
    """
    if include is None and exclude is None:

        def isvalid(tid):
            return True

    else:
        if include is None:
            include = set(tn.tensor_map)
        else:
            include = set(include)

        if exclude is None:
            exclude = set()
        else:
            exclude = set(exclude)

        def isvalid(tid):
            return (tid in include) and (tid not in exclude)

    d = 0
    if isinstance(tids, int):
        # allow single tid to be passed
        patch = {tids}
    else:
        patch = set(tids)

    boundary = tuple(patch)
    while d < max_distance:
        # expand outwards
        next_boundary = set()
        for tid in boundary:
            for tid_n in tn._get_neighbor_tids(tid):
                if isvalid(tid_n):
                    patch.add(tid_n)
                    next_boundary.add(tid_n)
        boundary = tuple(next_boundary)
        d += 1

    return tuple(sorted(patch))


def get_path_between_tids(tn, tida, tidb):
    """Find a shortest path between ``tida`` and ``tidb`` in this tensor
    network. Returns a ``NetworkPath`` if a path is found, otherwise ``None``.

    Currently ignores dangling and hyper indices.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to find a path in.
    tida : int
        The tensor id to start from.
    tidb : int
        The tensor id to end at.

    Returns
    -------
    NetworkPath or None
    """
    # expand from both points
    path_a0 = NetworkPath((tida,))
    path_b0 = NetworkPath((tidb,))
    queue_a = collections.deque((path_a0,))
    queue_b = collections.deque((path_b0,))
    # track ends of path so we identify when they meet
    # also acts a store for shortest path to that point
    ends_a = {tida: path_a0}
    ends_b = {tidb: path_b0}

    while queue_a or queue_b:
        for queue, ends_this, ends_other in [
            (queue_a, ends_a, ends_b),
            (queue_b, ends_b, ends_a),
        ]:
            if not queue:
                # no possible path
                return None

            path = queue.popleft()

            # get the tensor at the current end of the path
            last_tid = path.tids[-1]
            t = tn.tensor_map[last_tid]

            # check ways we could extend it
            for next_ind in t.inds:
                if next_ind in path:
                    # don't go back on ourselves
                    continue

                tids = tn.ind_map[next_ind]
                if len(tids) != 2:
                    # ignore dangling and hyper indices
                    continue

                next_tid = next(tid for tid in tids if tid != last_tid)

                if next_tid in ends_this:
                    # already been here in shorter or equal path
                    continue

                if next_tid in ends_other:
                    # found valid complete path!
                    other_path = ends_other[next_tid]

                    # want path to go from tida -> tidb
                    if queue is queue_a:
                        return NetworkPath(
                            tids=path.tids + other_path.tids[::-1],
                            inds=path.inds
                            + (next_ind,)
                            + other_path.inds[::-1],
                        )
                    else:
                        return NetworkPath(
                            tids=other_path.tids + path.tids[::-1],
                            inds=other_path.inds
                            + (next_ind,)
                            + path.inds[::-1],
                        )

                # valid partial path
                next_path = path.extend(next_ind, next_tid)
                ends_this[next_tid] = next_path
                queue.append(next_path)


def gen_all_paths_between_tids(tn, tida, tidb):
    """Generate all shortest paths between ``tida`` and ``tidb`` in this
    tensor network. Returns a generator of ``NetworkPath`` objects, ignores
    dangling and hyper indices currently.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to find paths in.
    tida : int
        The tensor id to start from.
    tidb : int
        The tensor id to end at.

    Yields
    ------
    NetworkPath
    """
    # map of only those neighbors which contribute to shortest paths
    predecessors = {}
    distances = {tidb: 0}
    queue = collections.deque([(tidb, 0)])
    found_start = False

    while queue:
        # get possible path extension, BFS
        last_tid, length = queue.popleft()

        # check ways we could extend it
        t = tn.tensor_map[last_tid]
        for next_ind in t.inds:
            tids = tn.ind_map[next_ind]
            if len(tids) != 2:
                # ignore dangling and hyper indices
                continue
            next_tid = next(tid for tid in tids if tid != last_tid)

            if next_tid == tida:
                found_start = True

            d = distances.get(next_tid, None)
            if d is None:
                # first time reaching this node
                distances[next_tid] = length + 1
                predecessors[next_tid] = [(last_tid, next_ind)]
                if not found_start:
                    # BFS search, so once we have found target, all
                    # possible paths will be in the queue already
                    queue.append((next_tid, length + 1))
            elif length < d:
                # another shortest path, just update predecessors
                # since extentions handled by case above
                predecessors[next_tid].append((last_tid, next_ind))

    # back track to find all paths
    queue = [NetworkPath([tida])]
    while queue:
        # this part can be DFS
        path = queue.pop()
        last_tid = path.tids[-1]
        for next_tid, next_ind in predecessors[last_tid]:
            new_path = path.extend(next_ind, next_tid)
            if next_tid == tidb:
                # reached the start
                yield new_path
            else:
                queue.append(new_path)


def gen_paths_loops(
    tn,
    max_loop_length=None,
    intersect=False,
    tids=None,
    inds=None,
    paths=None,
):
    """Generate all paths, up to a specified length, that represent loops in
    this tensor network. Unlike ``gen_loops`` this function will yield a
    `NetworkPath` objects, allowing one to differentiate between e.g. a double
    loop and a 'figure of eight' loop. Dangling and hyper indices are ignored.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to find loops in.
    max_loop_length : None or int
        Set the maximum number of indices that can appear in a loop. If
        ``None``, wait until any loop is found and set that as the maximum
        length.
    intersect : bool, optional
        Whether to allow self-intersecting loops.
    tids : None or sequence of int, optional
        If supplied, only consider loops containing one of these tensor ids.
    inds : None or sequence of str, optional
        If supplied, only consider loops containing one of these indices.
    paths : None or sequence of NetworkPath, optional
        If supplied, only consider loops starting from these paths.

    Yields
    ------
    NetworkPath

    See Also
    --------
    gen_loops, gen_inds_connected
    """
    queue = collections.deque()

    if isinstance(tids, int):
        # allow single tid to be passed
        tids = (tids,)
    if isinstance(inds, str):
        # allow single index to be passed
        inds = (inds,)

    if (tids is None) and (inds is None) and (paths is None):
        # default to finding loops everywhere
        inds = tn.ind_map

    if tids is not None:
        # generate loops starting at any of these tids
        for tid in tids:
            queue.append(NetworkPath([tid]))

    if inds is not None:
        # generate loops passing through any of these indices
        for ind in inds:
            tids = tn.ind_map[ind]
            if len(tids) != 2:
                # ignore dangling and hyper indices
                continue

            tida, tidb = tids
            # (only need one direction)
            queue.append(NetworkPath((tida, tidb), (ind,)))

    if paths is not None:
        # generate loops starting from these paths
        for path in paths:
            if not isinstance(path, NetworkPath):
                path = NetworkPath.from_sequence(path)
            queue.append(path)

    # cache index neighbor lookups for speed
    neighbormap = {}
    seen = set()

    while queue:
        path = queue.popleft()

        if intersect:
            # might have formed a closed loop, then it matter where we are
            # continuing from, so key on both ends
            search_key = (path.key, frozenset((path.tids[0], path.tids[-1])))
        else:
            # set of tids and inds is unique for non-intersecting loops
            search_key = path.key

        if search_key in seen:
            continue
        seen.add(search_key)

        last_tid = path.tids[-1]
        try:
            last_ind = path.inds[-1]
        except IndexError:
            # path is a single tid, no indices
            last_ind = None

        try:
            expansions = neighbormap[last_ind, last_tid]
        except KeyError:
            # check which ways we can continue this path
            possible_inds = tn.tensor_map[last_tid].inds
            expansions = []
            for next_ind in possible_inds:
                # don't come back the way we came
                if next_ind != last_ind:
                    next_ind_tids = tn.ind_map[next_ind]
                    # only consider normal bonds
                    if len(next_ind_tids) == 2:
                        # get the tid which isn't the direction we came
                        next_tid, next_tid_b = next_ind_tids
                        if next_tid == last_tid:
                            next_tid = next_tid_b
                        expansions.append((next_ind, next_tid))

            # cache this lookup
            neighbormap[last_ind, last_tid] = expansions

        continue_search = (max_loop_length is None) or (
            len(path) < max_loop_length - 1
        )

        for next_ind, next_tid in expansions:
            if next_ind in path:
                # can't ever double up on indices
                continue

            if next_tid == path.tids[0]:
                # finished a loop!

                loop = path.extend(next_ind, next_tid)
                if loop.key not in seen:
                    seen.add(loop.key)
                    if max_loop_length is None:
                        max_loop_length = len(loop)

                    # normalize the loop to be consistent across searches
                    # yield _normalize_loop(loop)
                    yield loop

            if continue_search and (intersect or next_tid not in path):
                # valid candidate extension!
                # -> we can double up on nodes only if intersecting
                queue.append(path.extend(next_ind, next_tid))


def gen_sloops(
    tn,
    max_loop_length=None,
    num_joins=1,
    intersect=False,
    tids=None,
    inds=None,
    paths=None,
):
    if num_joins < 1:
        return ()

    base_loops = gen_paths_loops(
        tn,
        max_loop_length=max_loop_length,
        intersect=intersect,
        tids=tids,
        inds=inds,
        paths=paths,
    )

    if num_joins == 1:
        # just return the base loops
        return base_loops

    # will reuse
    current_patches = tuple(base_loops)

    if (tids is None) and (inds is None) and (paths is None):
        # loops are already global
        base_loops = current_patches
    else:
        # need to merge local base loops with global loops
        # XXX: do this with the tids in current_patches, every join
        base_loops = tuple(
            gen_paths_loops(
                tn, max_loop_length=max_loop_length, intersect=intersect
            )
        )

    # efficient lookup of overlapping loop tids
    lookup = {}
    for sl in base_loops:
        for tid in unique(sl.tids):
            lookup.setdefault(tid, []).append(sl)

    for _ in range(num_joins - 1):
        next_patches = set()
        for p in current_patches:
            once = set()
            twice = set()

            # for each tensor
            for tid in unique(p.tids):
                # lookup possible base loops to merge with
                for po in lookup[tid]:
                    if po in once:
                        twice.add(po)
                    else:
                        once.add(po)

            for po in twice:
                # merge and add!
                next_patches.add(p.merge(po))

        current_patches = tuple(next_patches)

    return current_patches


def gen_patches(tn, max_size, tids=None, grow_from="all"):
    """Generate sets of tids that represent 'patches' of the tensor network,
    where each patch is a connected subgraph of the tensor network. Unlike
    generalized loops, patches can contain dangling nodes.
    """
    if tids is None:
        # find regions anywhere
        tids = tn.tensor_map.keys()
        grow_from = "any"
    elif isinstance(tids, int):
        # handle single tid region
        tids = (tids,)

    if grow_from == "all":
        # take `tids` as single initial region
        queue = collections.deque((frozenset(tids),))
    elif grow_from == "any":
        # take each tid as an initial region
        queue = collections.deque(frozenset([tid]) for tid in tids)
    else:
        raise ValueError("`grow_from` must be 'all' or 'any'.")

    # cache neighbors for speed
    neighbormap = {}
    seen = set()

    while queue:
        region = queue.popleft()

        if len(region) <= max_size:
            # is a valid patch
            yield tuple(sorted(region))

            if len(region) < max_size:
                # is valid to be extended
                # first get all atch neighbors
                outer = set()
                for tid in region:
                    try:
                        neighbors = neighbormap[tid]
                    except KeyError:
                        neighbors = tn._get_neighbor_tids([tid])
                        neighbormap[tid] = neighbors

                    for ntid in neighbors:
                        if ntid not in region:
                            outer.add(ntid)

                for ntid in outer:
                    # then extend this patch with each new combination
                    nregion = region | {ntid}
                    if nregion not in seen:
                        queue.append(nregion)
                        seen.add(nregion)


def _gen_gloops_single(tn, max_size=None, tids=None, grow_from="all"):
    if tids is None:
        # find loops everywhere
        tids = tn.tensor_map.keys()
        grow_from = "any"
    elif isinstance(tids, int):
        # handle single tid region
        tids = (tids,)

    if grow_from in ("all", "alldangle"):
        # take `tids` as single initial region
        queue = collections.deque((frozenset(tids),))
    elif grow_from in ("any", "anydangle"):
        # take each tid as an initial region
        queue = collections.deque(frozenset([tid]) for tid in tids)
    else:
        raise ValueError("`grow_from` must be 'all' or 'any'.")

    if "dangle" in grow_from:
        # target tids are allowed to be dangling
        dangle_tids = set(tids)
    else:
        dangle_tids = set()

    # cache neighbors for speed
    tid2inds = {}
    seen = set()

    while queue:
        region = queue.popleft()

        inds_once = set()
        inds_more = set()
        tids_next = set()

        for tid in region:
            try:
                inds = tid2inds[tid]
            except KeyError:
                inds = tid2inds[tid] = tn.tensor_map[tid].inds

            for ind in inds:
                if ind in inds_once:
                    # already seen -> is a bond
                    inds_more.add(ind)
                else:
                    inds_once.add(ind)

        valid_gloop = True
        for tid in region:
            # count number of connections each node has within region
            num_inner_connections = 0
            for ind in tid2inds[tid]:
                if ind in inds_more:
                    # bond within region
                    num_inner_connections += 1
                else:
                    # bond to outside -> add neighbors to queue
                    for ntid in tn.ind_map[ind]:
                        if ntid != tid:
                            tids_next.add(ntid)

            valid_gloop &= (
                # site is allowed to dangle
                tid in dangle_tids
                # or it is connected by at least two bonds
                or num_inner_connections >= 2
            )

        if valid_gloop:
            # valid region: no node is connected by a single bond only
            if max_size is None:
                # automatically set maximum region size
                max_size = len(region)
            yield tuple(sorted(region))

        if (max_size is None) or len(region) < max_size:
            # continue searching
            for ntid in tids_next:
                # possible extensions
                nregion = region | {ntid}
                if nregion not in seen:
                    # many ways to construct a region -> only check one
                    queue.append(nregion)
                    seen.add(nregion)


def gen_gloops(
    tn,
    max_size=None,
    tids=None,
    grow_from="all",
    num_joins=1,
    join_overlap=2,
):
    """Generate sets of tids that represent 'generalized loops' where every
    node is connected to at least two bonds, i.e. 2-degree connected subgraphs.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to find loops in.
    max_size : None or int
        Set the maximum number of tensors that can appear in a region. If
        ``None``, wait until any valid region is found and set that as the
        maximum size.
    tids : None or sequence of int, optional
        If supplied, only yield loops containing these tids, see
        ``grow_from``.
    grow_from : {'all', 'any', 'alldangle', 'anydangle'}, optional
        Only if ``tids`` is specified, this determines how to filter
        loops. If 'all', only yield loops containing *all* of the tids
        in ``tids``, if 'any', yield loops containing *any* of the tids
        in ``tids``. If 'alldangle' or 'anydangle', the tids are allowed to
        be dangling, i.e. 1-degree connected. This is useful for computing
        local expectations where the operator insertion breaks the loop
        assumption locally.
    num_joins : int, optional
        If larger than 1, repeatedly generate larger loops by joining together
        the initial set (individually with size up to ``max_size``) of
        generalized loops. Each join combines loops that overlap on at least
        ``join_overlap`` tids.
    join_overlap : {1, 2}, optional
        When joining loops together, the minimum number of overlapping
        tids they much share. 1 allows merging on a single node, 2 requires
        sharing a bond, which leads to fewer but 'denser' loops.

    Yields
    ------
    tuple[int]
    """
    if num_joins < 1:
        return ()

    base_gloops = _gen_gloops_single(
        tn,
        max_size=max_size,
        tids=tids,
        grow_from=grow_from,
    )

    if num_joins == 1:
        # just return the base loops
        return base_gloops

    # will reuse
    current_patches = tuple(map(frozenset, base_gloops))

    if tids is None:
        # loops are already global
        base_gloops = current_patches
    else:
        # need to merge local base loops with global loops
        # XXX: do this with the tids in current_patches, every join
        base_gloops = tuple(
            map(frozenset, _gen_gloops_single(tn, max_size=max_size))
        )

    # efficient lookup of overlapping gloop tids
    lookup = {}
    for gl in base_gloops:
        for tid in gl:
            lookup.setdefault(tid, []).append(gl)

    once = set()
    twice = set()

    if join_overlap == 1:
        overlapping = once
    elif join_overlap == 2:
        overlapping = twice
    else:
        raise ValueError("`join_overlap` must be 1 or 2.")

    for _ in range(num_joins - 1):
        next_patches = set()
        for gl in current_patches:
            once.clear()
            twice.clear()

            # for each tensor
            for tid in gl:
                # lookup possible base loops to merge with
                for glo in lookup[tid]:
                    if glo in once:
                        twice.add(glo)
                    else:
                        once.add(glo)

            for glo in overlapping:
                # merge and add!
                next_patches.add(gl | glo)

        current_patches = tuple(next_patches)

    return current_patches


def gen_loops(tn, max_loop_length=None):
    """Generate sequences of tids that represent loops in the TN.

    Parameters
    ----------
    max_loop_length : None or int
        Set the maximum number of tensors that can appear in a loop. If
        ``None``, wait until any loop is found and set that as the
        maximum length.

    Yields
    ------
    tuple[int]

    See Also
    --------
    gen_paths_loops
    """
    from cotengra.core import get_hypergraph

    inputs = {tid: t.inds for tid, t in tn.tensor_map.items()}
    hg = get_hypergraph(inputs, accel="auto")
    return hg.compute_loops(max_loop_length=max_loop_length)


def get_loop_union(
    tn,
    tids,
    max_size=None,
    grow_from="all",
):
    """Find the union, in terms of tids, of all generliazed loops that pass
    through either all or at least one of the given tids, depending on
    `grow_from`.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to find the loop union region in.
    tids : sequence of int
        The tensor ids to consider.
    max_size : None or int, optional
        The maximum number of tensors that can appear in the region. If
        ``None``, wait until any valid region is found and set that as the
        maximum size.
    grow_from : {'all', 'any', 'alldangle', 'anydangle'}, optional
        Only if ``tids`` is specified, this determines how to filter
        loops. If 'all', only take loops containing *all* of the tids
        in ``tids``, if 'any', yield loops containing *any* of the tids
        in ``tids``. If 'alldangle' or 'anydangle', the base tids are allowed
        to be dangling, i.e. 1-degree connected.

    Returns
    -------
    tuple[int]
    """
    gloops = _gen_gloops_single(
        tn,
        max_size=max_size,
        tids=tids,
        grow_from=grow_from,
    )
    return tuple(sorted({tid for r in gloops for tid in r}))


def gen_inds_connected(tn, max_length):
    """Generate all index 'patches' of size up to ``max_length``.

    Parameters
    ----------
    max_length : int
        The maximum number of indices in the patch.

    Yields
    ------
    tuple[str]

    See Also
    --------
    gen_paths_loops
    """
    queue = [(ix,) for ix in tn.ind_map]
    seen = {frozenset(s) for s in queue}
    while queue:
        s = queue.pop()
        if len(s) == max_length:
            continue
        expansions = tn._get_neighbor_inds(s)
        for ix in expansions:
            next_s = s + (ix,)
            key = frozenset(next_s)
            if key not in seen:
                # new string
                yield next_s
                seen.add(key)
                queue.append(next_s)


def tids_are_connected(tn, tids):
    """Check whether nodes ``tids`` are connected.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to check.
    tids : sequence of int
        Nodes to check.

    Returns
    -------
    bool
    """
    enum = range(len(tids))
    groups = dict(zip(enum, enum))
    regions = [(oset([tid]), tn._get_neighbor_tids(tid)) for tid in tids]
    for i, j in itertools.combinations(enum, 2):
        mi = groups.get(i, i)
        mj = groups.get(j, j)

        if regions[mi][0] & regions[mj][1]:
            groups[mj] = mi
            regions[mi][0].update(regions[mj][0])
            regions[mi][1].update(regions[mj][1])

    return len(set(groups.values())) == 1


def compute_shortest_distances(tn, tids=None, exclude_inds=()):
    """Compute the minimum graph distances between all or some nodes
    ``tids``.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to compute distances in.
    tids : None or sequence of int, optional
        If supplied, only compute distances between these nodes.
    exclude_inds : sequence of str, optional
        Exclude these indices when computing distances.

    Returns
    -------
    dict[tuple[int, int], int]
    """
    if tids is None:
        tids = tn.tensor_map
    else:
        tids = set(tids)

    visitors = collections.defaultdict(frozenset)
    for tid in tids:
        # start with target tids having 'visited' themselves only
        visitors[tid] = frozenset([tid])

    distances = {}
    N = math.comb(len(tids), 2)

    for d in itertools.count(1):
        any_change = False
        old_visitors = visitors.copy()

        # only need to iterate over touched region
        for tid in tuple(visitors):
            # at each step, each node sends its current visitors to all
            # neighboring nodes
            current_visitors = old_visitors[tid]
            for next_tid in tn._get_neighbor_tids(tid, exclude_inds):
                visitors[next_tid] |= current_visitors

        for tid in tuple(visitors):
            # check for new visitors -> those with shortest path d
            for diff_tid in visitors[tid] - old_visitors[tid]:
                any_change = True
                if (tid in tids) and (diff_tid in tids) and (tid < diff_tid):
                    distances[tid, diff_tid] = d

        if (len(distances) == N) or (not any_change):
            # all pair combinations have been computed, or everything
            # converged, presumably due to disconnected subgraphs
            break

    return distances


def compute_hierarchical_linkage(
    tn,
    tids=None,
    method="weighted",
    optimal_ordering=True,
    exclude_inds=(),
):
    from scipy.cluster import hierarchy

    if tids is None:
        tids = tn.tensor_map

    try:
        from cotengra import get_hypergraph

        hg = get_hypergraph(
            {tid: t.inds for tid, t in tn.tensor_map.items()},
            accel="auto",
        )
        for ix in exclude_inds:
            hg.remove_edge(ix)
        y = hg.all_shortest_distances_condensed(tuple(tids))
        return hierarchy.linkage(
            y, method=method, optimal_ordering=optimal_ordering
        )
    except ImportError:
        pass

    distances = tn.compute_shortest_distances(tids, exclude_inds)

    dinf = 10 * tn.num_tensors
    y = [
        distances.get(tuple(sorted((i, j))), dinf)
        for i, j in itertools.combinations(tids, 2)
    ]

    return hierarchy.linkage(
        y, method=method, optimal_ordering=optimal_ordering
    )


def compute_hierarchical_ssa_path(
    tn,
    tids=None,
    method="weighted",
    optimal_ordering=True,
    exclude_inds=(),
    are_sorted=False,
    linkage=None,
):
    """Compute a hierarchical grouping of ``tids``, as a ``ssa_path``."""
    if linkage is None:
        linkage = tn.compute_hierarchical_linkage(
            tids,
            method=method,
            exclude_inds=exclude_inds,
            optimal_ordering=optimal_ordering,
        )

    sorted_ssa_path = ((int(x[0]), int(x[1])) for x in linkage)
    if are_sorted:
        return tuple(sorted_ssa_path)

    if tids is None:
        tids = tn.tensor_map
    given_idx = {tid: i for i, tid in enumerate(tids)}
    sorted_to_given_idx = {
        i: given_idx[tid] for i, tid in enumerate(sorted(tids))
    }
    return tuple(
        (sorted_to_given_idx.get(x, x), sorted_to_given_idx.get(y, y))
        for x, y in sorted_ssa_path
    )


def compute_hierarchical_ordering(
    tn,
    tids=None,
    method="weighted",
    optimal_ordering=True,
    exclude_inds=(),
    linkage=None,
):
    """Compute a hierarchical ordering of ``tids``."""
    from scipy.cluster import hierarchy

    if tids is None:
        tids = list(tn.tensor_map)

    if linkage is None:
        linkage = tn.compute_hierarchical_linkage(
            tids,
            method=method,
            exclude_inds=exclude_inds,
            optimal_ordering=optimal_ordering,
        )

    node2tid = {i: tid for i, tid in enumerate(sorted(tids))}
    return tuple(map(node2tid.__getitem__, hierarchy.leaves_list(linkage)))


def compute_hierarchical_grouping(
    tn,
    max_group_size,
    tids=None,
    method="weighted",
    optimal_ordering=True,
    exclude_inds=(),
    linkage=None,
):
    """Group ``tids`` (by default, all tensors) into groups of size
    ``max_group_size`` or less, using a hierarchical clustering.
    """
    if tids is None:
        tids = list(tn.tensor_map)

    tids = sorted(tids)

    if linkage is None:
        linkage = tn.compute_hierarchical_linkage(
            tids,
            method=method,
            exclude_inds=exclude_inds,
            optimal_ordering=optimal_ordering,
        )

    ssa_path = tn.compute_hierarchical_ssa_path(
        tids=tids,
        method=method,
        exclude_inds=exclude_inds,
        are_sorted=True,
        linkage=linkage,
    )

    # follow ssa_path, agglomerating groups as long they small enough
    groups = {i: (tid,) for i, tid in enumerate(tids)}
    ssa = len(tids) - 1
    for i, j in ssa_path:
        ssa += 1

        if (i not in groups) or (j not in groups):
            # children already too big
            continue

        if len(groups[i]) + len(groups[j]) > max_group_size:
            # too big, skip
            continue

        # merge groups
        groups[ssa] = groups.pop(i) + groups.pop(j)

    # now sort groups by when their nodes in leaf ordering
    ordering = tn.compute_hierarchical_ordering(
        tids=tids,
        method=method,
        exclude_inds=exclude_inds,
        optimal_ordering=optimal_ordering,
        linkage=linkage,
    )
    score = {tid: i for i, tid in enumerate(ordering)}
    groups = sorted(
        groups.items(), key=lambda kv: sum(map(score.__getitem__, kv[1]))
    )

    return tuple(kv[1] for kv in groups)


def compute_centralities(tn):
    """Compute a simple centrality measure for each tensor in the network. The
    values go from 0 to 1, with 1 being the most central tensor.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to compute centralities for.

    Returns
    -------
    dict[int, float]
    """
    import cotengra as ctg

    hg = ctg.get_hypergraph({tid: t.inds for tid, t in tn.tensor_map.items()})
    return hg.simple_centrality()


def most_central_tid(tn):
    """Find the most central tensor in the network."""
    cents = tn.compute_centralities()
    return max((score, tid) for tid, score in cents.items())[1]


def least_central_tid(tn):
    """Find the least central tensor in the network."""
    cents = tn.compute_centralities()
    return min((score, tid) for tid, score in cents.items())[1]



"""Hyper dense belief propagation for arbitrary `quimb` tensor networks. This
is the classic 1-norm version of belief propagation, which treats the tensor
network directly as a factor graph. Messages are processed one at a time.

TODO:

- [ ] implement 'touching', so that only necessary messages are updated
- [ ] implement sequential update

"""

import autoray as ar

import quimb.tensor as qtn

from .bp_common import (
    BeliefPropagationCommon,
    combine_local_contractions,
    compute_all_index_marginals_from_messages,
    contract_hyper_messages,
    initialize_hyper_messages,
    prod,
)


def compute_all_hyperind_messages_prod(ms, smudge_factor=1e-12):
    """Given set of messages ``ms`` incident to a single index, compute the
    corresponding next output messages, using the 'product' implementation.
    """
    if len(ms) == 2:
        # shortcut for 2 messages
        return [ms[1], ms[0]]

    x = prod(ms)
    return [x / (m + smudge_factor) for m in ms]


def compute_all_hyperind_messages_tree(ms):
    """Given set of messages ``ms`` incident to a single index, compute the
    corresponding next output messages, using the 'tree' implementation.
    """
    ndim = len(ms)
    if len(ms) == 2:
        # shortcut for 2 messages
        return [ms[1], ms[0]]

    mouts = [None for _ in range(ndim)]
    queue = [(tuple(range(ndim)), 1, ms)]

    while queue:
        js, x, ms = queue.pop()

        ndim = len(ms)
        if ndim == 1:
            # reached single message
            mouts[js[0]] = x
            continue
        elif ndim == 2:
            # shortcut for 2 messages left
            mouts[js[0]] = x * ms[1]
            mouts[js[1]] = ms[0] * x
            continue

        # else split in two and contract each half
        k = ndim // 2
        jl, jr = js[:k], js[k:]
        ml, mr = ms[:k], ms[k:]

        # contract the right messages to get new left array
        xl = prod((*mr, x))

        # contract the left messages to get new right array
        xr = prod((*ml, x))

        # add the queue for possible further halving
        queue.append((jl, xl, ml))
        queue.append((jr, xr, mr))

    return mouts


def compute_all_tensor_messages_shortcuts(x, ms, ndim):
    if ndim == 2:
        # shortcut for 2 messages
        return [x @ ms[1], ms[0] @ x]
    elif ndim == 1:
        # shortcut for single message
        return [x]
    elif ndim == 0:
        # shortcut for no messages
        return []


def compute_all_tensor_messages_prod(
    x,
    ms,
    backend=None,
    smudge_factor=1e-12,
):
    """Given set of messages ``ms`` incident to tensor with data ``x``, compute
    the corresponding next output messages, using the 'prod' implementation.
    """
    ndim = len(ms)
    if ndim <= 2:
        return compute_all_tensor_messages_shortcuts(x, ms, ndim)

    js = tuple(range(ndim))

    mx = qtn.array_contract(
        arrays=(x, *ms), inputs=(js, *((j,) for j in js)), output=js
    )
    mouts = []

    for j, g in enumerate(ms):
        mouts.append(
            qtn.array_contract(
                arrays=(mx, 1 / (g + smudge_factor)),
                inputs=(js, (j,)),
                output=(j,),
                backend=backend,
            )
        )

    return mouts


def compute_all_tensor_messages_tree(x, ms, backend=None):
    """Given set of messages ``ms`` incident to tensor with data ``x``, compute
    the corresponding next output messages, using the 'tree' implementation.
    """
    ndim = len(ms)
    if ndim <= 2:
        return compute_all_tensor_messages_shortcuts(x, ms, ndim)

    mouts = [None for _ in range(ndim)]
    queue = [(tuple(range(ndim)), x, ms)]

    while queue:
        js, x, ms = queue.pop()

        ndim = len(ms)
        if ndim == 1:
            # reached single message
            mouts[js[0]] = x
            continue
        elif ndim == 2:
            # shortcut for 2 messages left
            mouts[js[0]] = x @ ms[1]
            mouts[js[1]] = ms[0] @ x
            continue

        # else split in two and contract each half
        k = ndim // 2
        jl, jr = js[:k], js[k:]
        ml, mr = ms[:k], ms[k:]

        # contract the right messages to get new left array
        xl = qtn.array_contract(
            arrays=(x, *mr),
            inputs=(js, *((j,) for j in jr)),
            output=jl,
            backend=backend,
        )

        # contract the left messages to get new right array
        xr = qtn.array_contract(
            arrays=(x, *ml),
            inputs=(js, *((j,) for j in jl)),
            output=jr,
            backend=backend,
        )

        # add the queue for possible further halving
        queue.append((jl, xl, ml))
        queue.append((jr, xr, mr))

    return mouts


class HD1BP(BeliefPropagationCommon):
    """Object interface for hyper, dense, 1-norm belief propagation. This is
    standard belief propagation in tensor network form.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to run BP on.
    messages : dict, optional
        Initial messages to use, if not given then uniform messages are used.
    damping : float or callable, optional
        The damping factor to apply to messages. This simply mixes some part
        of the old message into the new one, with the final message being
        ``damping * old + (1 - damping) * new``. This makes convergence more
        reliable but slower.
    update : {'sequential', 'parallel'}, optional
        Whether to update messages sequentially (newly computed messages are
        immediately used for other updates in the same iteration round) or in
        parallel (all messages are comptued using messages from the previous
        round only). Sequential generally helps convergence but parallel can
        possibly converge to differnt solutions.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    smudge_factor : float, optional
        A small number to add to the denominator of messages to avoid division
        by zero. Note when this happens the numerator will also be zero.
    inplace : bool, optional
        Whether to perform any operations inplace on the input tensor network.
    """

    def __init__(
        self,
        tn,
        *,
        messages=None,
        damping=0.0,
        update="sequential",
        normalize=None,
        distance=None,
        smudge_factor=1e-12,
        inplace=False,
    ):
        super().__init__(
            tn,
            damping=damping,
            update=update,
            normalize=normalize,
            distance=distance,
            inplace=inplace,
        )

        self.smudge_factor = smudge_factor

        if callable(messages):
            messages = initialize_hyper_messages(
                tn, fill_fn=messages, smudge_factor=smudge_factor
            )
        elif messages is None:
            messages = initialize_hyper_messages(
                tn, smudge_factor=smudge_factor
            )
        self.messages = messages

    def iterate(self, tol=None):
        if self.update == "sequential":
            new_messages = self.messages
        else:
            new_messages = {}

        def _normalize_and_insert(key, new_m, max_mdiff):
            # normalize and insert
            new_m = self._normalize_fn(new_m)
            old_m = self.messages[key]

            # pre-damp distance
            mdiff = self._distance_fn(old_m, new_m)

            if self.damping:
                new_m = self._damping_fn(old_m, new_m)

            # # post-damp distance
            # mdiff = self._distance_fn(old_m, new_m)

            max_mdiff = max(mdiff, max_mdiff)

            # set and return the max diff so far
            new_messages[key] = new_m
            return max_mdiff

        max_mdiff = 0.0

        # hyper index messages
        for ix, tids in self.tn.ind_map.items():
            ms = compute_all_hyperind_messages_prod(
                [self.messages[tid, ix] for tid in tids],
                self.smudge_factor,
            )
            for tid, m in zip(tids, ms):
                max_mdiff = _normalize_and_insert((ix, tid), m, max_mdiff)

        if self.update == "parallel":
            self.messages.update(new_messages)
            new_messages.clear()

        # tensor messages
        for tid, t in self.tn.tensor_map.items():
            inds = t.inds
            ms = compute_all_tensor_messages_tree(
                t.data,
                [self.messages[ix, tid] for ix in inds],
            )
            for ix, m in zip(inds, ms):
                max_mdiff = _normalize_and_insert((tid, ix), m, max_mdiff)

        if self.update == "parallel":
            self.messages.update(new_messages)

        return max_mdiff

    def get_gauged_tn(self):
        """Assuming the supplied tensor network has no hyper or dangling
        indices, gauge it by inserting the BP-approximated transfer matrix
        eigenvectors, which may be complex. The BP-contraction of this gauged
        network is then simply the product of zeroth entries of each tensor.
        """
        tng = self.tn.copy()
        for ind, tids in self.tn.ind_map.items():
            tida, tidb = tids
            ka = (ind, tida)
            kb = (ind, tidb)
            ma = self.messages[ka]
            mb = self.messages[kb]

            el, ev = ar.do("linalg.eig", ar.do("outer", ma, mb))
            k = ar.do("argsort", -ar.do("abs", el))
            ev = ev[:, k]
            Uinv = ev
            U = ar.do("linalg.inv", ev)
            tng._insert_gauge_tids(U, tida, tidb, Uinv)
        return tng

    def contract(self, strip_exponent=False, check_zero=True):
        """Estimate the total contraction, i.e. the exponential of the 'Bethe
        free entropy'.
        """
        return contract_hyper_messages(
            self.tn,
            self.messages,
            strip_exponent=strip_exponent,
            check_zero=check_zero,
            backend=self.backend,
        )

    def normalize_messages(self):
        """Normalize all messages such that the 'region contraction' of a
        single hyper index is 1.
        """
        for ind, tids in self.tn.ind_map.items():
            ms = [self.messages[tid, ind] for tid in tids]
            overlap = qtn.array_contract(ms, [(0,) for _ in ms], [])
            overlap **= 1 / len(ms)
            for tid, m in zip(tids, ms):
                self.messages[tid, ind] = m / overlap

    def get_cluster(self, r, virtual=True, autocomplete=True):
        """Get the tensor network of a region ``r``, with all boundary
        messages attached.

        Parameters
        ----------
        r : sequence of int or str
            The region to get, given as a sequence of indices or tensor ids.
        virtual : bool, optional
            Whether the view the original tensors (`virtual=True`, the default)
            or take copies (`virtual=False`).
        autocomplete : bool, optional
            Whether to automatically include all indices attached to the
            tensors in the region, or just the ones given in ``r``.

        Returns
        -------
        TensorNetwork
        """
        rtids = set()
        rinds = set()
        for x in r:
            if isinstance(x, int):
                rtids.add(x)
                if autocomplete:
                    rinds.update(self.tn.tensor_map[x].inds)
            else:
                rinds.add(x)

        tnr = self.tn._select_tids(rtids, virtual=virtual)
        for ind in rinds:
            # attach all messages coming from tensors outside the cluster
            for ntid in self.tn.ind_map[ind]:
                if ntid not in rtids:
                    tnr |= qtn.Tensor(
                        data=self.messages[ntid, ind], inds=(ind,)
                    )

        return tnr

    def contract_gloop_expand(
        self,
        gloops=None,
        strip_exponent=False,
        check_zero=True,
        optimize="auto-hq",
        progbar=False,
        **contract_otps,
    ):
        from .regions import gen_region_counts

        # if we normalized messages we can ignore all index-only regions
        self.normalize_messages()

        if isinstance(gloops, int):
            gloops = tuple(self.tn.gen_gloops(gloops))

        region_counts = gen_region_counts(gloops)

        if progbar:
            import tqdm

            region_counts = tqdm.tqdm(region_counts)

        zvals = []
        for r, cr in region_counts:
            # XXX: autoreduce intersecting clusters to gloops?
            # either we autocomplete above or we do it here per region
            tnr = self.get_cluster(r, autocomplete=True)
            zr = tnr.contract(
                output_inds=(), optimize=optimize, **contract_otps
            )
            zvals.append((zr, cr))

        return combine_local_contractions(
            zvals,
            strip_exponent=strip_exponent,
            check_zero=check_zero,
            backend=self.backend,
        )


def contract_hd1bp(
    tn,
    messages=None,
    max_iterations=1000,
    tol=5e-6,
    damping=0.0,
    diis=False,
    update="sequential",
    normalize=None,
    distance=None,
    tol_abs=None,
    tol_rolling_diff=None,
    smudge_factor=1e-12,
    strip_exponent=False,
    check_zero=True,
    info=None,
    progbar=False,
):
    """Estimate the contraction of ``tn`` with hyper, vectorized, 1-norm
    belief propagation, via the exponential of the Bethe free entropy.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to run BP on, can have hyper indices.
    messages : dict, optional
        Initial messages to use, if not given then uniform messages are used.
    max_iterations : int, optional
        The maximum number of iterations to perform.
    tol : float, optional
        The convergence tolerance for messages.
    damping : float, optional
        The damping parameter to use, defaults to no damping.
    diis : bool or dict, optional
        Whether to use direct inversion in the iterative subspace to
        help converge the messages by extrapolating to low error guesses.
        If a dict, should contain options for the DIIS algorithm. The
        relevant options are {`max_history`, `beta`, `rcond`}.
    update : {'sequential', 'parallel'}, optional
        Whether to update messages sequentially or in parallel.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    tol_abs : float, optional
        The absolute convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``.
    tol_rolling_diff : float, optional
        The rolling mean convergence tolerance for maximum message update
        distance, if not given then taken as ``tol``. This is used to stop
        running when the messages are just bouncing around the same level,
        without any overall upward or downward trends, roughly speaking.
    smudge_factor : float, optional
        A small number to add to the denominator of messages to avoid division
        by zero. Note when this happens the numerator will also be zero.
    strip_exponent : bool, optional
        Whether to strip the exponent from the final result. If ``True``
        then the returned result is ``(mantissa, exponent)``.
    check_zero : bool, optional
        Whether to check for zero values and return zero early.
    info : dict, optional
        If specified, update this dictionary with information about the
        belief propagation run.
    progbar : bool, optional
        Whether to show a progress bar.

    Returns
    -------
    scalar or (scalar, float)
    """
    bp = HD1BP(
        tn,
        messages=messages,
        damping=damping,
        update=update,
        normalize=normalize,
        distance=distance,
        smudge_factor=smudge_factor,
    )
    bp.run(
        max_iterations=max_iterations,
        tol=tol,
        diis=diis,
        tol_abs=tol_abs,
        tol_rolling_diff=tol_rolling_diff,
        info=info,
        progbar=progbar,
    )
    return bp.contract(
        strip_exponent=strip_exponent,
        check_zero=check_zero,
    )


def run_belief_propagation_hd1bp(
    tn,
    messages=None,
    max_iterations=1000,
    tol=5e-6,
    damping=0.0,
    smudge_factor=1e-12,
    info=None,
    progbar=False,
):
    """Run belief propagation on a tensor network until it converges. This
    is the basic version that does not vectorize contractions.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to run BP on.
    messages : dict, optional
        The current messages. For every index and tensor id pair, there should
        be a message to and from with keys ``(ix, tid)`` and ``(tid, ix)``.
        If not given, then messages are initialized as uniform.
    max_iterations : int, optional
        The maximum number of iterations to run for.
    tol : float, optional
        The convergence tolerance.
    smudge_factor : float, optional
        A small number to add to the denominator of messages to avoid division
        by zero. Note when this happens the numerator will also be zero.
    info : dict, optional
        If specified, update this dictionary with information about the
        belief propagation run.
    progbar : bool, optional
        Whether to show a progress bar.

    Returns
    -------
    messages : dict
        The final messages.
    converged : bool
        Whether the algorithm converged.
    """
    bp = HD1BP(
        tn, messages=messages, damping=damping, smudge_factor=smudge_factor
    )
    bp.run(max_iterations=max_iterations, tol=tol, info=info, progbar=progbar)
    return bp.messages, bp.converged


def sample_hd1bp(
    tn,
    messages=None,
    output_inds=None,
    max_iterations=1000,
    tol=1e-2,
    damping=0.0,
    smudge_factor=1e-12,
    bias=False,
    seed=None,
    progbar=False,
):
    """Sample all indices of a tensor network using repeated belief propagation
    runs and decimation.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to sample.
    messages : dict, optional
        The current messages. For every index and tensor id pair, there should
        be a message to and from with keys ``(ix, tid)`` and ``(tid, ix)``.
        If not given, then messages are initialized as uniform.
    output_inds : sequence of str, optional
        The indices to sample. If not given, then all indices are sampled.
    max_iterations : int, optional
        The maximum number of iterations for each message passing run.
    tol : float, optional
        The convergence tolerance for each message passing run.
    smudge_factor : float, optional
        A small number to add to each message to avoid zeros. Making this large
        is similar to adding a temperature, which can aid convergence but
        likely produces less accurate marginals.
    bias : bool or float, optional
        Whether to bias the sampling towards the largest marginal. If ``False``
        (the default), then indices are sampled proportional to their
        marginals. If ``True``, then each index is 'sampled' to be its largest
        weight value always. If a float, then the local probability
        distribution is raised to this power before sampling.
    thread_pool : bool, int or ThreadPoolExecutor, optional
        Whether to use a thread pool for parallelization. If an integer, then
        this is the number of threads to use. If ``True``, then the number of
        threads is set to the number of cores. If a ``ThreadPoolExecutor``,
        then this is used directly.
    seed : int, optional
        A random seed to use for the sampling.
    progbar : bool, optional
        Whether to show a progress bar.

    Returns
    -------
    config : dict[str, int]
        The sample configuration, mapping indices to values.
    tn_config : TensorNetwork
        The tensor network with all index values (or just those in
        `output_inds` if supllied) selected. Contracting this tensor network
        (which will just be a sequence of scalars if all index values have been
        sampled) gives the weight of the sample, e.g. should be 1 for a SAT
        problem and valid assignment.
    omega : float
        The probability of choosing this sample (i.e. product of marginal
        values). Useful possibly for importance sampling.
    """
    import numpy as np

    rng = np.random.default_rng(seed)

    tn_config = tn.copy()

    if messages is None:
        messages = initialize_hyper_messages(tn_config)

    if output_inds is None:
        output_inds = tn_config.ind_map.keys()
    output_inds = set(output_inds)

    config = {}
    omega = 1.0

    if progbar:
        import tqdm

        pbar = tqdm.tqdm(total=len(output_inds))
    else:
        pbar = None

    while output_inds:
        messages, _ = run_belief_propagation_hd1bp(
            tn_config,
            messages,
            max_iterations=max_iterations,
            tol=tol,
            damping=damping,
            smudge_factor=smudge_factor,
            progbar=True,
        )

        marginals = compute_all_index_marginals_from_messages(
            tn_config, messages
        )

        # choose most peaked marginal
        ix, p = max(
            (m for m in marginals.items() if m[0] in output_inds),
            key=lambda ix_p: max(ix_p[1]),
        )

        if bias is False:
            # sample the value according to the marginal
            v = rng.choice(np.arange(p.size), p=p)
        elif bias is True:
            v = np.argmax(p)
            # in some sense omega is really 1.0 here
        else:
            # bias towards larger marginals by raising to a power
            p = p**bias
            p /= np.sum(p)
            v = np.random.choice(np.arange(p.size), p=p)

        omega *= p[v]
        config[ix] = v

        # clean up messages
        for tid in tn_config.ind_map[ix]:
            del messages[ix, tid]
            del messages[tid, ix]

        # remove index
        tn_config.isel_({ix: v})
        output_inds.remove(ix)

        if progbar:
            pbar.update(1)
            pbar.set_description(f"{ix}->{v}", refresh=False)

    if progbar:
        pbar.close()

    return config, tn_config, omega



"""Implementation of arbitrary geometry wavefunction cluster update."""

import functools

from quimb.tensor.tensor_core import ensure_dict, bonds, bonds_size, do
from quimb.tensor.tensor_arbgeom_tebd import SimpleUpdateGen


def gate_inds_nn_fit(
    self,
    G,
    ind1,
    ind2,
    max_bond=None,
    method="als",
    pregauge=2,
    init_simple_guess=True,
    steps=10,
    fit_opts=None,
    contract_opts=None,
    inplace=False,
):
    """Gate two nearest neighbor outer indices, using full fitting of
    reduced tensors with respect to the environment. This is more accurate
    than a simple reduced gate when restricting the bond dimension.

    Parameters
    ----------
    G : array_like
        The gate to fit.
    ind1, ind2 : str
        The indices to gate.
    max_bond : int, optional
        The maximum bond dimension to use. If ``None``, use the maximum
        bond dimension that the tensors currently share.
    method : {'als', 'autodiff'}, optional
        The method to use for fitting.
    pregauge : int, optional
        How many times to locally canonize from the purified environment
        tensor to both the left and right reduced tensors.
    init_simple_guess : bool, optional
        Whether to use a 'simple update' guess for the initial guess. This
        can be quite high quality already if pregauging is used.
    steps : int, optional
        The number of steps to use for fitting, can be ``0`` in which case
        the initial guess is used, which in conjuction with the
        envinronment pregauging can still be quite high quality.
    inplace : bool, optional
        Whether to update the tensor network in place.
    contract_opts
        Supplied to
        :meth:`~quimb.tensor.tensor_core.TensorNetwork.contract`.
    """
    fit_opts = ensure_dict(fit_opts)
    contract_opts = ensure_dict(contract_opts)

    ket = self.copy()

    # move indices onto shared bond so environment can be contracted
    ket.reduce_inds_onto_bond(
        ind1, ind2, combine=False, ndim_cutoff=0, tags="__REDUCED__"
    )

    ket.add_tag("__KET__")
    bra = ket.conj().retag_({"__KET__": "__BRA__"})
    norm = ket | bra

    # contract environment -> all but reduced tensors
    norm.contract_tags_("__REDUCED__", "!any", **contract_opts)

    # get tensors and bond names
    (tide,) = norm._get_tids_from_tags("__REDUCED__", "!any")
    (te,) = norm._tids_get(tide)
    (tid_k1,) = norm._get_tids_from_inds(ind1) & norm._get_tids_from_tags(
        "__KET__"
    )
    (tid_k2,) = norm._get_tids_from_inds(ind2) & norm._get_tids_from_tags(
        "__KET__"
    )
    (tid_b1,) = norm._get_tids_from_inds(ind1) & norm._get_tids_from_tags(
        "__BRA__"
    )
    (tid_b2,) = norm._get_tids_from_inds(ind2) & norm._get_tids_from_tags(
        "__BRA__"
    )
    tk1 = norm.tensor_map[tid_k1]
    tk2 = norm.tensor_map[tid_k2]
    tb1 = norm.tensor_map[tid_b1]
    tb2 = norm.tensor_map[tid_b2]

    (ix_ek1,) = bonds(te, tk1)
    (ix_ek2,) = bonds(te, tk2)
    (ix_eb1,) = bonds(te, tb1)
    (ix_eb2,) = bonds(te, tb2)

    if max_bond is None:
        max_bond = bonds_size(tk1, tk2)

    # split environment to get purification
    _, tek = te.split(
        left_inds=[ix_eb1, ix_eb2],
        right_inds=[ix_ek1, ix_ek2],
        method="svd",
    )

    #   ┌────────────────┐
    # ┌─┴┐      ┌──┐     │   :
    # │k1├──────┤k2├───┐ │   : this is the purification
    # └─┬┘      └─┬┘   ├─┴┐  : we'll fit to gated version of itself
    #   │         │    │ek│  :
    #   │ind1     │    └─┬┘
    #   │     ind2│    ┌─┴┐
    #   │         │    │eb│
    # ┌─┴┐      ┌─┴┐   ├─┬┘
    # │b1├──────┤b2├───┘ │
    # └─┬┘      └──┘     │
    #   └────────────────┘

    for _ in range(int(pregauge)):
        # perform some conditioning: local gauging from Q -> R tensors
        for ind, ix_ek, tk in [
            (ind1, ix_ek1, tk1),
            (ind2, ix_ek2, tk2),
        ]:
            R = tek.split(
                left_inds=None,
                right_inds=[ix_ek],
                method="qr",
                get="arrays",
            )[1]

            Rinv = do("linalg.inv", R)
            # get Q tensor in uncontracted TN
            (tidkq,) = ket._get_tids_from_inds(
                ix_ek
            ) - ket._get_tids_from_inds(ind)
            tkq = ket.tensor_map[tidkq]
            # need to keep environment and bare tensors in sync
            tkq.gate_(Rinv.T, ix_ek)
            tek.gate_(Rinv.T, ix_ek)
            tk.gate_(R, ix_ek)

    # form purification
    tnl = tek | tk1 | tk2
    # form gated purification
    tnl_target = tnl.gate_inds(G, (ind1, ind2), contract=True)

    if init_simple_guess:
        # maybe initialize with simple guess
        tnl.gate_inds_(
            G, (ind1, ind2), contract="split", max_bond=max_bond, cutoff=0.0
        )

    if steps:
        # perform the actual fitting, specifying only the reduced tensors
        tnl.fit_(
            tnl_target,
            tags=["__REDUCED__"],
            method=method,
            steps=steps,
            **fit_opts,
        )

    # re-absorb reduced factors
    ket.contract_ind(ix_ek1)
    ket.contract_ind(ix_ek2)

    # TN we will return
    new = self if inplace else self.copy()
    (tn1,) = new._inds_get(ind1)
    (tn2,) = new._inds_get(ind2)

    # permute to match original tensors
    (t1,) = ket._inds_get(ind1)
    (t2,) = ket._inds_get(ind2)
    t1.transpose_like_(tn1)
    t2.transpose_like_(tn2)

    tn1.modify(data=t1.data)
    tn2.modify(data=t2.data)

    return new


gate_inds_nn_fit_ = functools.partialmethod(gate_inds_nn_fit, inplace=True)

# TensorNetwork.gate_inds_nn_fit =gate_inds_nn_fit
# TensorNetwork.gate_inds_nn_fit_ = gate_inds_nn_fit


class ClusterUpdateNNGen(SimpleUpdateGen):
    """Cluster update for arbitrary geometry nearest neighbor hamiltonians.
    This keeps track of simple update style gauges, in order to approximately
    partial trace beyond ``cluster_radius`` and form an approximate environment
    for two nearest neighbor sites that be used to fit the gate with higher
    quality than simple update only.
    """

    def __init__(
        self,
        psi0,
        ham,
        tau=0.01,
        D=None,
        cluster_radius=1,
        cluster_fillin=0,
        gauge_smudge=1e-6,
        imag=True,
        gate_opts=None,
        ordering=None,
        second_order_reflect=False,
        compute_energy_every=None,
        compute_energy_final=True,
        compute_energy_opts=None,
        compute_energy_fn=None,
        compute_energy_per_site=False,
        callback=None,
        keep_best=False,
        progbar=True,
    ):
        super().__init__(
            psi0=psi0,
            ham=ham,
            tau=tau,
            D=D,
            imag=imag,
            gate_opts=gate_opts,
            ordering=ordering,
            second_order_reflect=second_order_reflect,
            compute_energy_every=compute_energy_every,
            compute_energy_final=compute_energy_final,
            compute_energy_opts=compute_energy_opts,
            compute_energy_fn=compute_energy_fn,
            compute_energy_per_site=compute_energy_per_site,
            callback=callback,
            keep_best=keep_best,
            progbar=progbar,
        )
        self.cluster_radius = cluster_radius
        self.cluster_fillin = cluster_fillin
        self.gauge_smudge = gauge_smudge

        # override some default TEBDGen gate_opts
        self.gate_opts.pop("cutoff")
        self.gate_opts.pop("contract")

    def gate(self, U, where):
        taga, tagb = self._psi.gen_tags_from_coos(where)
        inda, indb = self._psi.gen_inds_from_coos(where)

        # get the local cluster
        psi_local = self._psi.select_local(
            (taga, tagb),
            "any",
            max_distance=self.cluster_radius,
            fillin=self.cluster_fillin,
            virtual=True,
        )

        # temporarily gauge it with 'simple' gauges
        with psi_local.gauge_simple_temp(
            self._gauges,
            smudge=self.gauge_smudge,
        ):
            # fit the gate to the gauged local cluster
            gate_inds_nn_fit(
                psi_local,
                U,
                inda,
                indb,
                **self.gate_opts,
            )

        # update nearest gauges for the modified tensors
        self._psi.gauge_local_(
            (taga, tagb),
            "any",
            max_distance=1,
            method="simple",
            gauges=self._gauges,
            smudge=self.gauge_smudge,
        )

        # perform some conditioning
        self._psi.equalize_norms_(1.0)
        self._psi.exponent = 0.0



"""Belief propagation (BP) routines. There are three potential categorizations
of BP and each combination of them is potentially valid specific algorithm.

1-norm vs 2-norm BP
-------------------

- 1-norm (normal): BP runs directly on the tensor network, messages have size
  ``d`` where ``d`` is the size of the bond(s) connecting two tensors or
  regions.
- 2-norm (quantum): BP runs on the squared tensor network, messages have size
  ``d^2`` where ``d`` is the size of the bond(s) connecting two tensors or
  regions. Each local tensor or region is partially traced (over dangling
  indices) with its conjugate to create a single node.


Graph vs Hypergraph BP
----------------------

- Graph (simple): the tensor network lives on a graph, where indices either
  appear on two tensors (a bond), or appear on a single tensor (are outputs).
  In this case, messages are exchanged directly between tensors.
- Hypergraph: the tensor network lives on a hypergraph, where indices can
  appear on any number of tensors. In this case, the update procedure is two
  parts, first all 'tensor' messages are computed, these are then used in the
  second step to compute all the 'index' messages, which are then fed back into
  the 'tensor' message update and so forth. For 2-norm BP one likely needs to
  specify which indices are outputs and should be traced over.

The hypergraph case of course includes the graph case, but since the 'index'
message update is simply the identity, it is convenient to have a separate
simpler implementation, where the standard TN bond vs physical index
definitions hold.


Dense vs Vectorized vs Lazy BP
------------------------------

- Dense: each node is a single tensor, or pair of tensors for 2-norm BP. If all
  multibonds have been fused, then each message is a vector (1-norm case) or
  matrix (2-norm case).
- Vectorized: the same as the above, but all matching tensor update and message
  updates are stacked and performed simultaneously. This can be enormously more
  efficient for large numbers of small tensors.
- Lazy: each node is potentially a tensor network itself with arbitrary inner
  structure and number of bonds connecting to other nodes. The message are
  generally tensors and each update is a lazy contraction, which is potentially
  much cheaper / requires less memory than forming the 'dense' node for large
  tensors.

(There is also the MPS flavor where each node has a 1D structure and the
messages are matrix product states, with updates involving compression.)


Overall that gives 12 possible BP flavors, some implemented here:

- [x] (HD1BP) hyper, dense, 1-norm - this is the standard BP algorithm
- [x] (HD2BP) hyper, dense, 2-norm
- [x] (HV1BP) hyper, vectorized, 1-norm
- [ ] (HV2BP) hyper, vectorized, 2-norm
- [ ] (HL1BP) hyper, lazy, 1-norm
- [ ] (HL2BP) hyper, lazy, 2-norm
- [x] (D1BP) simple, dense, 1-norm - simple BP for simple tensor networks
- [x] (D2BP) simple, dense, 2-norm - this is the standard PEPS BP algorithm
- [ ] (V1BP) simple, vectorized, 1-norm
- [ ] (V2BP) simple, vectorized, 2-norm
- [x] (L1BP) simple, lazy, 1-norm
- [x] (L2BP) simple, lazy, 2-norm

The 2-norm methods can be used to compress bonds or estimate the 2-norm.
The 1-norm methods can be used to estimate the 1-norm, i.e. contracted value.
Both methods can be used to compute index marginals and thus perform sampling.

The vectorized methods can be extremely fast for large numbers of small
tensors, but do currently require all dimensions to match.

The dense and lazy methods can can converge messages *locally*, i.e. only
update messages adjacent to messages which have changed.
"""

from .bp_common import combine_local_contractions, initialize_hyper_messages
from .d1bp import D1BP, contract_d1bp
from .d2bp import D2BP, compress_d2bp, contract_d2bp, sample_d2bp
from .hd1bp import HD1BP, contract_hd1bp, sample_hd1bp
from .hv1bp import HV1BP, contract_hv1bp, sample_hv1bp
from .l1bp import L1BP, contract_l1bp
from .l2bp import L2BP, compress_l2bp, contract_l2bp
from .regions import RegionGraph, gen_region_counts

__all__ = (
    "combine_local_contractions",
    "compress_d2bp",
    "compress_l2bp",
    "contract_d1bp",
    "contract_d2bp",
    "contract_hd1bp",
    "contract_hv1bp",
    "contract_l1bp",
    "contract_l2bp",
    "D1BP",
    "D2BP",
    "gen_region_counts",
    "HD1BP",
    "HV1BP",
    "initialize_hyper_messages",
    "L1BP",
    "L2BP",
    "RegionGraph",
    "sample_d2bp",
    "sample_hd1bp",
    "sample_hv1bp",
)



import numpy as np

from quimb.tensor import Tensor, TensorNetwork, tensor_contract
from quimb.tensor.belief_propagation import (
    RegionGraph,
    combine_local_contractions,
)
from quimb.tensor.belief_propagation.bp_common import BeliefPropagationCommon


def auto_add_indices(tn, regions):
    """Make sure all indices incident to any tensor in each region are
    included in the region.
    """
    new_regions = []
    for r in regions:
        new_r = set(r)
        tids = [x for x in new_r if isinstance(x, int)]
        for tid in tids:
            t = tn.tensor_map[tid]
            new_r.update(t.inds)
        new_regions.append(frozenset(new_r))
    return new_regions


class HD1GBP(BeliefPropagationCommon):
    """Generalized belief propagation for hyper tensor networks.

    Parameters
    ----------
    tn : TensorNetwork
        The hyper tensor network to run GBP on.
    regions : sequence[sequence[int | str]]
        The regions to use for GBP. Each region can be a set of tids and
        indices. If a tid is present in a region, all its indices are
        automatically included in the region when ``autocomplete=True``.
    autocomplete : bool, optional
        Whether to automatically compute all intersection subregions for the
        RegionGraph.
    autoprune : bool, optional
        Whether to automatically remove all regions with a count of zero.
    damping : float, optional
        The damping factor to use for the messages.
    optimize : str, optional
        The contraction path optimization strategy to use.
    """

    def __init__(
        self,
        tn: TensorNetwork,
        regions,
        *,
        messages=None,
        autocomplete=True,
        autoprune=True,
        damping=1 / 2,
        optimize="auto-hq",
        **kwargs,
    ):
        super().__init__(
            tn,
            damping=damping,
            **kwargs,
        )

        if autocomplete:
            regions = auto_add_indices(tn, regions)

        self.rg = RegionGraph(
            regions,
            autocomplete=autocomplete,
            autoprune=autoprune,
        )

        if callable(messages):
            self._message_init_function = messages
            self.messages = {}
        elif messages is not None:
            self._message_init_function = None
            self.messages = messages
        else:
            self._message_init_function = None
            self.messages = {}

        self.new_messages = {}
        self.contract_opts = dict(
            optimize=optimize,
            drop_tags=True,
        )

    def get_message_tensors(self, source, target):
        """Get all message tensors needed to compute the message from
        ``source`` to ``target``.
        """
        # get the nodes and edge keys for the message
        r_a_without_b, pairs_mul, pairs_div = self.rg.get_message_parts(
            (source, target)
        )

        ts = []
        # first add factors not in target region
        for x in r_a_without_b:
            if isinstance(x, int):
                ts.append(self.tn.tensor_map[x])

        # then messages only appearing in source belief
        for pair in pairs_mul:
            try:
                ts.append(self.messages[pair])
            except KeyError:
                pass

        # then messages only appearing in target belief
        # note we use the *new* messages here, as per GBP stability
        for pair in pairs_div:
            try:
                ts.append(1 / self.new_messages[pair])
            except KeyError:
                try:
                    ts.append(1 / self.messages[pair])
                except KeyError:
                    pass

        return ts

    def compute_message(
        self,
        source,
        target,
        **contract_opts,
    ):
        """Compute the message from source to target region.

        Parameters
        ----------
        source : Region
            The source region.
        target : Region
            The target region.
        contract_opts
            Supplied to :func:`~quimb.tensor.tensor_contract`.
        """
        contract_opts = {**self.contract_opts, **contract_opts}

        ts = self.get_message_tensors(source, target)

        if ts:
            # can only output indices which are present
            output_inds = sorted(
                {ind for t in ts for ind in t.inds}.intersection(target)
            )
            # perform the message contraction!
            m = tensor_contract(
                *ts,
                output_inds=output_inds,
                preserve_tensor=True,
                **contract_opts,
            )
        else:
            # output uniform distribution
            m = Tensor()

        # normalize
        m.modify(apply=self._normalize_fn)

        return m

    def iterate(self, tol=5e-6):
        max_mdiff = 0.0

        ncheck = 0
        nconv = 0

        # compute messages into smaller regions first
        for child in sorted(self.rg.regions, key=len):
            for parent in self.rg.get_parents(child):
                # contract the message!
                m = self.compute_message(parent, child)

                if self._message_init_function is not None:
                    # the messages change size during the first few iterations
                    # so we perform a delayed initialization upon seeing
                    # each fresh shape
                    mprev = self.new_messages.get((parent, child), None)
                    if (mprev is None) or (mprev.shape != m.shape):
                        m.modify(data=self._message_init_function(m.shape))

                # immediately update the new messages for stability in GBP
                # (they are used in the 'denominator' of higher messages)
                self.new_messages[parent, child] = m

                # check for convergence
                try:
                    m_old = self.messages[parent, child]
                    # XXX: need to handle index alignment here to compare
                    # using _distance_fn:
                    # mdiff = self._distance_fn(m_old.data, m.data)
                    mdiff = (m_old - m).norm()
                except KeyError:
                    mdiff = 1.0
                max_mdiff = max(mdiff, max_mdiff)

                ncheck += 1
                if mdiff < tol:
                    nconv += 1

        # damped update of messages
        #     note that the raw, undamped `new_messages` are used in the
        #     denominator of the message computations, and so kept 'as is'
        for pair in self.new_messages:
            if pair not in self.messages:
                # no old message yet
                self.messages[pair] = self.new_messages[pair]
            else:
                self.messages[pair] = self._damping_fn(
                    self.messages[pair],
                    self.new_messages[pair],
                )

        # self.new_messages.clear()

        return {
            "nconv": nconv,
            "ncheck": ncheck,
            "max_mdiff": max_mdiff,
        }

    def get_belief_tensors(self, region):
        """Get the (uncontracted) tensors for the belief of ``region``."""
        ts = []

        # add factors
        for x in region:
            if isinstance(x, int):
                ts.append(self.tn.tensor_map[x])

        # add messages
        for pair in self.rg.get_coparent_pairs(region):
            try:
                ts.append(self.messages[pair])
            except KeyError:
                pass

        return ts

    def contract(self, strip_exponent=False, check_zero=True):
        """Contract this tensor network given the current GBP messages.

        Parameters
        ----------
        sstrip_exponent : bool, optional
            Whether to strip the exponent from the final result. If ``True``
            then the returned result is ``(mantissa, exponent)``.

        Returns
        -------
        result : float or (float, float)
            The combined value, or the mantissa and exponent separately.
        """
        zvals = []

        for r in self.rg.regions:
            c = self.rg.get_count(r)
            ts = self.get_belief_tensors(r)
            if ts:
                zr = tensor_contract(*ts, output_inds=(), **self.contract_opts)
                zvals.append((zr, c))

        return combine_local_contractions(
            zvals,
            backend=self.backend,
            strip_exponent=strip_exponent,
            check_zero=check_zero,
        )

    def draw(self, rhighlight=None, zfactor=2):
        from quimb.schematic import Drawing

        tid2site = {}
        for site in self.tn.sites:
            (tid,) = self.tn._get_tids_from_tags(site)
            tid2site[tid] = site

        def region_to_site(region):
            z = self.rg.get_level(region) * zfactor + np.random.uniform(0, 0.2)
            tids = []
            for x in region:
                if isinstance(x, int):
                    tids.append(x)
                else:
                    tids.extend(self.tn.ind_map[x])

            sites = [tid2site[tid] for tid in tids]
            xs, ys = zip(*sites)
            xmean = sum(xs) / len(sites)
            ymean = sum(ys) / len(sites)
            return xmean, ymean, z

        def ind_to_pos(ind):
            tids = self.tn.ind_map[ind]
            sites = [tid2site[tid] for tid in tids]
            xs, ys = zip(*sites)
            xmean = sum(xs) / len(sites)
            ymean = sum(ys) / len(sites)
            return xmean, ymean

        def region_to_pos(region):
            z = self.rg.get_level(region) * zfactor + np.random.uniform(0, 0.2)
            poss = []
            for x in region:
                if isinstance(x, int):
                    poss.append(tid2site[x])
                else:
                    poss.append(ind_to_pos(x))
            return tuple((*pos, z) for pos in poss)

        d = Drawing(figsize=(10, 10))

        if rhighlight == "random":
            import random

            rhighlight = random.choice(self.rg.regions)

        if rhighlight is not None:
            rchildren = self.rg.get_children(rhighlight)
            rdescendents = self.rg.get_descendents(rhighlight)
            rparents = self.rg.get_parents(rhighlight)
            rcoparents = [x[0] for x in self.rg.get_coparent_pairs(rhighlight)]
            rancestors = self.rg.get_ancestors(rhighlight)
        else:
            rchildren = rdescendents = rparents = rcoparents = rancestors = []

        for r in self.rg.regions:
            # color = hash_to_color(str(r))
            # color = "grey"

            if r == rhighlight:
                color = (1.0, 0.0, 0.0, 0.3)
            elif r in rchildren:
                color = (1.0, 0.5, 0.0, 0.3)
            elif r in rdescendents:
                color = (1.0, 1.0, 0.0, 0.3)
            elif r in rparents:
                color = (0.2, 0.5, 0.8, 0.3)
            elif r in rcoparents:
                color = (0.3, 0.7, 0.5, 0.3)
            elif r in rancestors:
                color = (0.3, 0.5, 0.2, 0.3)
            else:
                color = (0.5, 0.5, 0.5, 0.1)

            pos = region_to_site(r)
            d.circle(pos, radius=0.05, color=color)
            tids = [x for x in r if isinstance(x, int)]
            if tids:
                d.patch_around(
                    region_to_pos(r),
                    radius=0.05,
                    facecolor=color,
                )

            for rc in self.rg.get_children(r):
                posc = region_to_site(rc)
                d.line(
                    pos,
                    posc,
                    color=color,
                    # arrowhead=False,
                )

        return d.fig, d.ax



import functools

import autoray as ar

import quimb.tensor as qtn
from quimb.tensor.tensor_core import check_opt
from quimb.tensor.tensor_2d import Rotator2D


def CTMRG(
    T,
    A,
    C,
    L,
    max_bond,
    strip_exponent=False,
):
    """Contract the translationally invariant tensor network given by bulk,
    edge and corner arrays ``T``, ``A`` and ``C`` respectively with side
    length ``L``.

    Parameters
    ----------
    T : array_like with shape (d, d, d, d)
        The bulk tensor, dimensions corresponding to
        (towards corner, away from corner, away from corner, towards corner),
    A : array_like with shape (d, d, d)
        The edge tensor, dimensions corresponding to
        (away from corner, towards corner, towards bulk).
    C : array_like with shape (d, d)
        The corner tensor.
    L : int
        The side length of the lattice to generate.
    max_bond : int
        The maximum bond dimension to use.
    strip_exponent : bool, optional
        Whether to strip an overall exponent while contracting and accrue this
        separately, to avoid overflow.

    Returns
    -------
    Z : float
        The contracted value of the tensor network.
    exponent : float, optional
        The exponent, in log10, accrued while contracting, such that the real
        contracted value is ``Z * 10**exponent``.
    """

    asymmetric = isinstance(A, tuple)
    if asymmetric:
        A, B = A
    else:
        B = A
        # # A = A
        # # perm = (1, 0, 2)
        # perm = (0, 1, 2)
        # B = A.transpose(perm)

    if (L < 4) or (L % 2 != 0):
        raise ValueError("`L` must be an even integer >= 4.")

    if strip_exponent:
        exponent = 0.0
    else:
        exponent = None

    order_a_inwards = True
    order_b_inwards = True

    C_inds = ("c-a", "c-b")
    if order_a_inwards:
        A_inds = ("c-a", "a-d", "a-t")
    else:
        A_inds = ("a-d", "c-a", "a-t")
    if order_b_inwards:
        B_inds = ("c-b", "b-r", "b-t")
    else:
        B_inds = ("b-r", "c-b", "b-t")
    T_inds = ("a-t", "t-d", "t-r", "b-t")

    AC_inds = ("a-d", "t-d")
    CB_inds = ("b-r", "t-r")

    d = T.shape[0]
    for l in range(2, L - 2, 2):
        #     ┌──┐      ┌──┐
        #     │C ├─c-b──┤B ├──b-r
        #     └┬─┘      └┬─┘
        #      │c-a      │b-t
        #     ┌┴─┐      ┌┴─┐
        #     │A ├─a-t──┤T ├──t-r
        #     └┬─┘      └┬─┘
        #      │a-d      │t-d
        tn_corner = (
            qtn.Tensor(C, inds=C_inds, tags="C")
            | qtn.Tensor(A, inds=A_inds, tags="A")
            | qtn.Tensor(B, inds=B_inds, tags="B")
            | qtn.Tensor(T, inds=T_inds, tags="T")
        )

        # make the projector
        CTM = tn_corner.to_dense(AC_inds, CB_inds)

        if asymmetric:
            s, U = ar.do("linalg.eigh", (CTM + ar.dag(CTM) / 2))
        else:
            # assert qu.isherm(CTM)
            s, U = ar.do("linalg.eigh", CTM)

        k = ar.do("argsort", -ar.do("abs", s))[:max_bond]
        U = U[:, k]
        s = s[k]
        U = U.reshape((-1, d, s.size))

        if asymmetric:
            tn_corner_proj = tn_corner.copy()
            tn_corner_proj |= qtn.Tensor(U, inds=(*AC_inds, "new-d"), tags="U")
            tn_corner_proj |= qtn.Tensor(U, inds=(*CB_inds, "new-r"), tags="V")
            C = tn_corner_proj.to_dense(["new-d"], ["new-r"])
        else:
            C = ar.do("diag", s)

        #      │a-u     │t-u
        #     ┌┴─┐     ┌┴─┐
        #     │A ├─a-t─┤T ├─t-r
        #     └┬─┘     └┬─┘
        #      │a-d     │t-d
        tn_side = (
            qtn.Tensor(A, inds=["a-u", "a-d", "a-t"], tags="A")
            if order_a_inwards
            else qtn.Tensor(A, inds=["a-d", "a-u", "a-t"], tags="A")
        ) | qtn.Tensor(T, inds=["a-t", "t-d", "t-r", "t-u"], tags="T")
        tn_side_proj = tn_side.copy()
        tn_side_proj |= qtn.Tensor(U, inds=["a-u", "t-u", "new-u"], tags="U")
        tn_side_proj |= qtn.Tensor(U, inds=["a-d", "t-d", "new-d"], tags="V")
        if order_a_inwards:
            A = tn_side_proj.to_dense(["new-u"], ["new-d"], ["t-r"])
        else:
            A = tn_side_proj.to_dense(["new-d"], ["new-u"], ["t-r"])

        if not asymmetric:
            B = A
        else:
            #          ┌──┐
            #      b-l─┤B ├──b-r
            #          └┬─┘
            #           │b-t
            #          ┌┴─┐
            #      t-l─┤T ├──t-r
            #          └┬─┘
            #           │t-d
            tn_side = (
                qtn.Tensor(B, inds=["b-l", "b-r", "b-t"], tags="B")
                if order_b_inwards
                else qtn.Tensor(B, inds=["b-r", "b-l", "b-t"], tags="B")
            ) | qtn.Tensor(T, inds=["t-l", "t-d", "t-r", "b-t"], tags="T")
            tn_side_proj = tn_side.copy()
            tn_side_proj |= qtn.Tensor(U, inds=["b-l", "t-l", "n-l"], tags="U")
            tn_side_proj |= qtn.Tensor(U, inds=["b-r", "t-r", "n-r"], tags="V")

            if order_b_inwards:
                B = tn_side_proj.to_dense(["n-l"], ["n-r"], ["t-d"])
            else:
                B = tn_side_proj.to_dense(["n-r"], ["n-l"], ["t-d"])

        if strip_exponent:
            # Anorm = ar.do('linalg.norm', A)
            # TODO: work out how to rescale given a different norm each time
            Anorm = 2
            exponent += 4 * l * ar.do("log10", Anorm)
            A = A / Anorm
            B = B / Anorm

            Cnorm = ar.do("linalg.norm", C)
            exponent += 4 * ar.do("log10", Cnorm)
            C = C / Cnorm

    tn_corner = (
        qtn.Tensor(C, inds=C_inds, tags="C")
        | qtn.Tensor(A, inds=A_inds, tags="A")
        | qtn.Tensor(B, inds=B_inds, tags="B")
        | qtn.Tensor(T, inds=T_inds, tags="T")
    )

    # make the projector
    CTM = tn_corner.to_dense(AC_inds, CB_inds)

    return ar.do("trace", CTM @ CTM @ CTM @ CTM), exponent


def coarse_grain_eager(
    self,
    direction,
    compress=True,
    equalize_norms=False,
    inplace=False,
    **compress_opts,
):
    """This contracts pairs of tensors in along ``direction``, and then
    optionally compresses the doubled bonds generated along the other
    direction.
    """
    check_opt("direction", direction, ("x", "y"))
    tn = self if inplace else self.copy()
    r2d = Rotator2D(tn, None, None, direction + "min", stepsize=2)

    # track new coordinates / tags
    retag_map = {}

    # first contract pairs along direction 'x' or 'y'
    for i in r2d.sweep:
        for j in r2d.sweep_other:
            #      │  │  │        │  │  │
            # i+1  │  O──O─       │  │  O─
            #      │╱ │  │   ->   │  │╱ │
            #  i  ═O──O──O─      ═O══O──O─
            #      │  │  │        │  │  │
            #         j              j
            tag_ij = r2d.site_tag(i, j)
            tag_ip1j = r2d.site_tag(i + 1, j)
            tn.contract_between(
                tag_ij, tag_ip1j, equalize_norms=equalize_norms
            )
            new_tag = r2d.site_tag(i // 2, j)
            retag_map[tag_ij] = new_tag
            retag_map[tag_ip1j] = new_tag

            if compress and j > 0:
                #      │  │  │        │  │  │
                # i+1  │  │  O─       │  │  O─
                #      │  │╱ │   ->   │  │╱ │
                #  i  ─O══O──O─      ─O──O──O─
                #      │  │  │        │  │  │
                #    j-1  j         j-1  j
                tag_ijm1 = r2d.site_tag(i, j - 1)
                tn.compress_between(tag_ijm1, tag_ij, **compress_opts)

        retag_map[r2d.x_tag(i)] = r2d.x_tag(i // 2)
        retag_map[r2d.x_tag(i + 1)] = r2d.x_tag(i // 2)

    # then we retag the tensor network and adjust its size
    tn.retag_(retag_map)
    if direction == "x":
        tn._Lx = tn.Lx // 2 + tn.Lx % 2
    else:
        tn._Ly = tn.Ly // 2 + tn.Ly % 2

    return tn


coarse_grain_ = functools.partialmethod(coarse_grain_eager, inplace=True)



import functools
import importlib

import numpy as np
import pytest
from autoray import real
from numpy.testing import assert_allclose

import quimb as qu
import quimb.tensor as qtn
from quimb.tensor.optimize import Vectorizer, parse_network_to_backend

found_torch = importlib.util.find_spec("torch") is not None
found_autograd = importlib.util.find_spec("autograd") is not None
found_jax = importlib.util.find_spec("jax") is not None
found_tensorflow = importlib.util.find_spec("tensorflow") is not None

if found_tensorflow:
    import tensorflow.experimental.numpy as tnp

    tnp.experimental_enable_numpy_behavior()

jax_case = pytest.param(
    "jax", marks=pytest.mark.skipif(not found_jax, reason="jax not installed")
)
autograd_case = pytest.param(
    "autograd",
    marks=pytest.mark.skipif(
        not found_autograd, reason="autograd not installed"
    ),
)
tensorflow_case = pytest.param(
    "tensorflow",
    marks=pytest.mark.skipif(
        not found_tensorflow, reason="tensorflow not installed"
    ),
)
pytorch_case = pytest.param(
    "torch",
    marks=pytest.mark.skipif(not found_torch, reason="pytorch not installed"),
)


@pytest.fixture
def tagged_qaoa_tn():
    """
    make qaoa tensor network, with RZZ and RX tagged on a per-round basis
    so that these tags can be used as shared_tags to TNOptimizer
    """

    n = 8
    depth = 4
    terms = [(i, (i + 1) % n) for i in range(n)]
    gammas = qu.randn(depth)
    betas = qu.randn(depth)

    # make circuit
    circuit_opts = {"gate_opts": {"contract": False}}
    circ = qtn.Circuit(n, **circuit_opts)

    # layer of hadamards to get into plus state
    for i in range(n):
        circ.apply_gate("H", i, gate_round=0)

    for d in range(depth):
        for i, j in terms:
            circ.apply_gate(
                "RZZ", -gammas[d], i, j, gate_round=d, parametrize=True
            )

        for i in range(n):
            circ.apply_gate(
                "RX", betas[d] * 2, i, gate_round=d, parametrize=True
            )

    # tag circuit for shared_tags
    tn_tagged = circ.psi.copy()
    for i in range(depth):
        tn_tagged.select(["RZZ", f"ROUND_{i}"]).add_tag(f"p{2 * i}")
        tn_tagged.select(["RX", f"ROUND_{i}"]).add_tag(f"p{2 * i + 1}")

    return n, depth, tn_tagged


@pytest.fixture
def heis_pbc():
    L = 10
    chi = 8
    dtype = "float32"
    psi0 = qtn.MPS_rand_state(L, chi, cyclic=True, seed=42).astype(dtype)
    H = qtn.MPO_ham_heis(L, cyclic=True).astype(dtype)

    def norm_fn(psi):
        factor = (psi & psi).contract(all, optimize="auto-hq")
        return psi / factor**0.5

    def loss_fn(psi, H):
        k, H, b = qtn.tensor_network_align(psi, H, psi)
        energy = (k & H & b).contract(all, optimize="auto-hq")
        return energy

    en_ex = qu.groundenergy(qu.ham_heis(L, cyclic=True, sparse=True))

    return psi0, H, norm_fn, loss_fn, en_ex


@pytest.fixture
def ham_mbl_pbc_complex():
    L = 10
    chi = 8
    dtype = "complex64"
    psi0 = qtn.MPS_rand_state(L, chi, cyclic=True, seed=42).astype(dtype)

    ham_opts = {"cyclic": True, "dh": 0.7, "dh_dim": 3, "seed": 42}
    H = qtn.MPO_ham_mbl(L, **ham_opts).astype(dtype)

    def norm_fn(psi):
        factor = (psi.H & psi).contract(all, optimize="auto-hq")
        return psi * factor**-0.5

    def loss_fn(psi, H):
        k, H, b = qtn.tensor_network_align(psi, H, psi.H)
        energy = (k & H & b).contract(all, optimize="auto-hq")
        return real(energy)

    en_ex = qu.groundenergy(qu.ham_mbl(L, sparse=True, **ham_opts))

    return psi0, H, norm_fn, loss_fn, en_ex


def test_vectorizer():
    shapes = [(2, 3), (4, 5), (6, 7, 8)]
    dtypes = ["complex64", "float32", "complex64"]
    arrays = [qu.randn(s, dtype=dtype) for s, dtype in zip(shapes, dtypes)]

    v = Vectorizer(arrays)

    grads = [qu.randn(s, dtype=dtype) for s, dtype in zip(shapes, dtypes)]
    v.pack(grads, "grad")

    new_arrays = v.unpack(v.vector)
    for x, y in zip(arrays, new_arrays):
        assert_allclose(x, y)

    new_arrays = v.unpack(v.grad)
    for x, y in zip(grads, new_arrays):
        assert_allclose(x, y)


def rand_array(rng):
    ndim = rng.integers(1, 6)
    shape = rng.integers(2, 6, size=ndim)
    dtype = rng.choice(["float32", "float64", "complex64", "complex128"])
    x = rng.normal(shape).astype(dtype)
    if "complex" in dtype:
        x += 1j * rng.normal(shape).astype(dtype)
    return x


def random_array_pytree(rng, max_depth=3):
    def _recurse(d=0):
        if d >= max_depth:
            return rand_array(rng)
        t = rng.choice(["array", "list", "tuple", "dict"])
        if t == "array":
            return rand_array(rng)
        elif t == "list":
            return [_recurse(d + 1) for _ in range(rng.integers(2, 6))]
        elif t == "tuple":
            return tuple([_recurse(d + 1) for _ in range(rng.integers(2, 6))])
        elif t == "dict":
            cs = (chr(i) for i in range(ord("a"), ord("z") + 1))
            return {
                next(cs): _recurse(d + 1) for _ in range(rng.integers(2, 6))
            }

    return _recurse()


def test_vectorizer_pytree():
    tree = random_array_pytree(np.random.default_rng(666))
    v = Vectorizer(tree)
    new_tree = v.unpack()
    assert tree is not new_tree
    assert str(tree) == str(new_tree)


@pytest.mark.parametrize(
    "backend", [jax_case, autograd_case, tensorflow_case, pytorch_case]
)
@pytest.mark.parametrize("method", ["simple", "basin"])
def test_optimize_pbc_heis(heis_pbc, backend, method):
    psi0, H, norm_fn, loss_fn, en_ex = heis_pbc
    tnopt = qtn.TNOptimizer(
        psi0,
        loss_fn,
        norm_fn,
        loss_constants={"H": H},
        autodiff_backend=backend,
    )
    if method == "simple":
        psi_opt = tnopt.optimize(100)
    elif method == "basin":
        psi_opt = tnopt.optimize_basinhopping(25, 4)
    assert loss_fn(psi_opt, H) == pytest.approx(en_ex, rel=1e-2)


@pytest.mark.parametrize(
    "backend", [jax_case, autograd_case, tensorflow_case, pytorch_case]
)
@pytest.mark.parametrize("method", ["simple", "basin"])
def test_optimize_ham_mbl_complex(ham_mbl_pbc_complex, backend, method):
    psi0, H, norm_fn, loss_fn, en_ex = ham_mbl_pbc_complex
    tnopt = qtn.TNOptimizer(
        psi0,
        loss_fn,
        norm_fn,
        loss_constants={"H": H},
        autodiff_backend=backend,
    )
    if method == "simple":
        psi_opt = tnopt.optimize(100)
    elif method == "basin":
        psi_opt = tnopt.optimize_basinhopping(25, 4)
    assert loss_fn(psi_opt, H) == pytest.approx(en_ex, rel=1e-2)


@pytest.mark.parametrize(
    "backend", [jax_case, autograd_case, pytorch_case, tensorflow_case]
)
def test_every_parametrized_gate(backend):
    circ = qtn.Circuit(2)
    circ.rx(*qu.randn(1), 0, parametrize=True, tags=["OPTIMIZE"])
    circ.ry(*qu.randn(1), 1, parametrize=True, tags=["OPTIMIZE"])
    circ.rz(*qu.randn(1), 0, parametrize=True, tags=["OPTIMIZE"])
    circ.rxx(*qu.randn(1), 0, 1, parametrize=True, tags=["OPTIMIZE"])
    circ.ryy(*qu.randn(1), 0, 1, parametrize=True, tags=["OPTIMIZE"])
    circ.rzz(*qu.randn(1), 1, 0, parametrize=True, tags=["OPTIMIZE"])
    circ.crx(*qu.randn(1), 1, 0, parametrize=True, tags=["OPTIMIZE"])
    circ.cry(*qu.randn(1), 0, 1, parametrize=True, tags=["OPTIMIZE"])
    circ.crz(*qu.randn(1), 1, 0, parametrize=True, tags=["OPTIMIZE"])
    circ.u1(*qu.randn(1), 1, parametrize=True, tags=["OPTIMIZE"])
    circ.u2(*qu.randn(2), 0, parametrize=True, tags=["OPTIMIZE"])
    circ.u3(*qu.randn(3), 1, parametrize=True, tags=["OPTIMIZE"])
    circ.cu1(*qu.randn(1), 0, 1, parametrize=True, tags=["OPTIMIZE"])
    circ.cu2(*qu.randn(2), 1, 0, parametrize=True, tags=["OPTIMIZE"])
    circ.cu3(*qu.randn(3), 1, 0, parametrize=True, tags=["OPTIMIZE"])
    circ.fsim(*qu.randn(2), 0, 1, parametrize=True, tags=["OPTIMIZE"])
    circ.fsimg(*qu.randn(5), 1, 0, parametrize=True, tags=["OPTIMIZE"])
    circ.givens(*qu.randn(1), 0, 1, parametrize=True, tags=["OPTIMIZE"])
    circ.givens2(*qu.randn(2), 0, 1, parametrize=True, tags=["OPTIMIZE"])
    circ.xx_plus_yy(*qu.randn(2), 0, 1, parametrize=True, tags=["OPTIMIZE"])
    circ.xx_minus_yy(*qu.randn(2), 0, 1, parametrize=True, tags=["OPTIMIZE"])
    circ.su4(*qu.randn(15), 0, 1, parametrize=True, tags=["OPTIMIZE"])
    psi = circ.psi

    target = qtn.Dense1D(qu.rand_haar_state(4))

    def loss(psi, target):
        return -(real(psi.H @ target) ** 2)

    f0 = loss(psi, target)
    tnopt = qtn.TNOptimizer(
        psi,
        loss,
        tags="OPTIMIZE",
        loss_constants={"target": target},
        autodiff_backend=backend,
        device="cpu",
    )
    tnopt.optimize(1)
    assert tnopt.loss < f0


@pytest.mark.parametrize(
    "backend", [jax_case, autograd_case, tensorflow_case, pytorch_case]
)
def test_parametrized_circuit(backend):
    H = qu.ham_mbl(4, dh=3.0, dh_dim=3)
    gs = qu.groundstate(H)
    T_gs = qtn.Dense1D(gs)

    def loss(psi, target):
        f = psi.H & target
        f.rank_simplify_()
        return -abs(f ^ all)

    circ = qtn.circ_ansatz_1D_brickwork(4, depth=4)
    psi0 = circ.psi
    tnopt = qtn.TNOptimizer(
        psi0,
        loss,
        tags="U3",
        loss_constants=dict(target=T_gs),
        autodiff_backend=backend,
        loss_target=-0.99,
    )
    psi_opt = tnopt.optimize(20)
    assert sum(loss < -0.99 for loss in tnopt.losses) == 1
    assert qu.fidelity(psi_opt.to_dense(), gs) > 0.99


def mera_norm_fn(mera):
    return mera.isometrize(method="cayley")


def mera_local_expectation(mera, terms, where):
    tags = [mera.site_tag(coo) for coo in where]
    mera_ij = mera.select(tags, "any")
    mera_ij_G = mera_ij.gate(terms[where], where)
    mera_ij_ex = mera_ij_G & mera_ij.H
    return mera_ij_ex.contract(all, optimize="auto-hq")


@pytest.mark.parametrize(
    "backend", [autograd_case, jax_case, tensorflow_case, pytorch_case]
)
@pytest.mark.parametrize("executor", [None, "threads"])
def test_multiloss(backend, executor):
    if executor == "threads":
        from concurrent.futures import ThreadPoolExecutor

        executor = ThreadPoolExecutor(2)

    L = 8
    D = 3
    dtype = "float32"

    mera = qtn.MERA.rand(L, max_bond=D, dtype=dtype)

    H2 = qu.ham_heis(2).real.astype(dtype)
    terms = {(i, (i + 1) % L): H2 for i in range(L)}

    loss_fns = [
        functools.partial(mera_local_expectation, where=where)
        for where in terms
    ]

    tnopt = qtn.TNOptimizer(
        mera,
        loss_fn=loss_fns,
        norm_fn=mera_norm_fn,
        loss_constants={"terms": terms},
        autodiff_backend=backend,
        executor=executor,
        device="cpu",
    )

    tnopt.optimize(10)
    # ex = -3.6510934089371734
    assert tnopt.loss < -2.5

    if executor is not None:
        executor.shutdown()


def test_parse_network_to_backend_shared_tags(tagged_qaoa_tn):
    n, depth, psi0 = tagged_qaoa_tn

    def to_constant(x):
        return np.asarray(x)

    tags = [f"p{i}" for i in range(2 * depth)]
    tn_tagged, variabes = parse_network_to_backend(
        psi0,
        tags=tags,
        shared_tags=tags,
        to_constant=to_constant,
    )
    # test number of variables identified
    assert len(variabes) == 2 * depth
    # each variable tag should be in n tensors
    for i in range(len(tags)):
        var_tag = f"__VARIABLE{i}__"
        assert len(tn_tagged.select(var_tag).tensors) == n


def test_parse_network_to_backend_individual_tags(tagged_qaoa_tn):
    n, depth, psi0 = tagged_qaoa_tn

    def to_constant(x):
        return np.asarray(x)

    tags = [f"p{i}" for i in range(2 * depth)]
    tn_tagged, variabes = parse_network_to_backend(
        psi0, tags=tags, to_constant=to_constant
    )
    # test number of variables identified
    assert len(variabes) == 2 * depth * n
    # each variable tag should only be in 1 tensors
    for i in range(len(tags)):
        var_tag = f"__VARIABLE{i}__"
        assert len(tn_tagged.select_tensors(var_tag)) == 1


def test_parse_network_to_backend_constant_tags(tagged_qaoa_tn):
    n, depth, psi0 = tagged_qaoa_tn

    def to_constant(x):
        return np.asarray(x)

    # constant tags, include shared variable tags for first QAOA layer
    constant_tags = ["PSI0", "H", "p0", "p1"]
    tn_tagged, variabes = parse_network_to_backend(
        psi0, constant_tags=constant_tags, to_constant=to_constant
    )

    # test number of variables identified
    assert len(variabes) == 2 * (depth - 1) * n
    # each variable tag should only be in 1 tensors
    for i in range(len(variabes)):
        var_tag = f"__VARIABLE{i}__"
        assert len(tn_tagged.select(var_tag).tensors) == 1


@pytest.mark.parametrize("backend", [jax_case, autograd_case, tensorflow_case])
def test_shared_tags(tagged_qaoa_tn, backend):
    n, depth, psi0 = tagged_qaoa_tn

    H = qu.ham_heis(
        n,
        j=(0.0, 0.0, -1.0),
        b=(1.0, 0.0, 0.0),
        cyclic=True,
    )
    gs = qu.groundstate(H)
    T_gs = qtn.Dense1D(gs).astype(complex)  # tensorflow needs all same dtype

    def loss(psi, target):
        f = psi.H & target
        f.rank_simplify_()
        return -abs(f ^ all)

    tags = [f"p{i}" for i in range(2 * depth)]
    tnopt = qtn.TNOptimizer(
        psi0,
        loss_fn=loss,
        tags=tags,
        shared_tags=tags,
        loss_constants={"target": T_gs},
        autodiff_backend=backend,
        # loss_target=-0.99,
    )

    # run optimisation and test output
    psi_opt = tnopt.optimize_basinhopping(n=10, nhop=5)
    # assert sum(loss < -0.99 for loss in tnopt.losses) == 1
    assert qu.fidelity(psi_opt.to_dense(), gs) > 0.99

    # test dimension of optimisation space
    assert tnopt.res.x.size == 2 * depth

    # examine tensors inside optimised TN and check sharing was done
    for tag in tags:
        test_data = None
        for t in psi_opt.select_tensors(tag):
            if test_data is None:
                test_data = t.get_params()
            else:
                assert_allclose(test_data, t.get_params())


@pytest.mark.parametrize(
    "backend", [jax_case, autograd_case, tensorflow_case, pytorch_case]
)
@pytest.mark.parametrize("simplify", ["ADCRS", "R"])
def test_optimize_circuit_directly(backend, simplify):
    if backend == "jax" and simplify == "ADCRS":
        pytest.skip("JAX does not support dynamic simplification.")

    circ = qtn.Circuit(2)
    rng = np.random.default_rng(42)
    circ.u3(*rng.uniform(high=2 * np.pi, size=3), 0, parametrize=True)
    circ.u3(*rng.uniform(high=2 * np.pi, size=3), 1, parametrize=True)
    circ.cnot(0, 1)
    circ.u3(*rng.uniform(high=2 * np.pi, size=3), 0, parametrize=True)
    circ.u3(*rng.uniform(high=2 * np.pi, size=3), 1, parametrize=True)

    H = qu.ham_heis(2).astype("complex128")

    def loss(circ, H):
        return real(
            circ.local_expectation(H, (0, 1), simplify_sequence=simplify)
        )

    assert loss(circ, H) > -0.74
    tnopt = qtn.TNOptimizer(
        circ, loss, loss_constants=dict(H=H), autodiff_backend=backend
    )
    circ_opt = tnopt.optimize(10)
    assert circ_opt is not circ
    assert loss(circ_opt, H) < -0.74
    assert {t.backend for t in circ_opt.psi} == {"numpy"}



import importlib
import itertools

import pytest

import quimb as qu
import quimb.tensor as qtn

found_torch = importlib.util.find_spec("torch") is not None

pytorch_case = pytest.param(
    "torch",
    marks=pytest.mark.skipif(not found_torch, reason="pytorch not installed"),
)


class TestLocalHam2DConstruct:
    @pytest.mark.parametrize("H2_type", ["default", "manual"])
    @pytest.mark.parametrize("H1_type", [None, "default", "manual"])
    @pytest.mark.parametrize("Lx", [3, 4])
    @pytest.mark.parametrize("Ly", [3, 4])
    def test_construct(self, Lx, Ly, H2_type, H1_type):
        import matplotlib
        from matplotlib import pyplot as plt

        matplotlib.use("Template")

        if H2_type == "default":
            H2 = qu.rand_herm(4)
        elif H2_type == "manual":
            H2 = dict()
            for i, j in itertools.product(range(Lx), range(Ly)):
                if i + 1 < Lx:
                    H2[(i, j), (i + 1, j)] = qu.rand_herm(4)
                if j + 1 < Ly:
                    H2[(i, j), (i, j + 1)] = qu.rand_herm(4)

        if H1_type is None:
            H1 = None
        elif H1_type == "default":
            H1 = qu.rand_herm(2)
        elif H1_type == "manual":
            H1 = dict()
            for i, j in itertools.product(range(Lx), range(Ly)):
                H1[i, j] = qu.rand_herm(2)

        ham = qtn.LocalHam2D(Lx, Ly, H2, H1)
        assert len(ham.terms) == 2 * Lx * Ly - Lx - Ly

        # check that terms are being cached if possible
        if (H2_type == "default") and (H1_type is None):
            assert len({id(x) for x in ham.terms.values()}) == 1

        print(ham)
        fig, ax = ham.draw()
        plt.close(fig)

    @pytest.mark.parametrize("Lx", [4, 5])
    @pytest.mark.parametrize("Ly", [4, 5])
    @pytest.mark.parametrize(
        "order", [None, "sort", "random", "smallest_last"]
    )
    def test_ordering(self, Lx, Ly, order):
        ham = qtn.ham_2d_j1j2(Lx, Ly)
        assert len(ham.terms) == 2 * Lx * Ly - Lx - Ly + 2 * (Lx - 1) * (
            Ly - 1
        )
        ordering = ham.get_auto_ordering(order)
        assert len(ordering) == len(ham.terms)
        assert set(ordering) == set(ham.terms)
        assert tuple(ordering) != tuple(ham.terms)

        # make sure first four pairs are in same commuting group at least
        first_four_pairs = tuple(itertools.chain(*ordering[:4]))
        assert len(first_four_pairs) == len(set(first_four_pairs))


class TestSimpleUpdate:
    @pytest.mark.parametrize("backend", ["numpy", pytorch_case])
    def test_heis_small(self, backend):
        Lx = 3
        Ly = 4
        D = 2

        ham = qtn.ham_2d_heis(Lx, Ly)
        psi0 = qtn.PEPS.rand(Lx, Ly, D, seed=42)

        def to_backend(x):
            import autoray

            return autoray.do("array", x, like=backend)

        psi0.apply_to_arrays(to_backend)
        ham.apply_to_arrays(to_backend)

        su = qtn.SimpleUpdate(
            psi0,
            ham,
            progbar=True,
            keep_best=True,
            compute_energy_every=10,
            ordering="largest_first",
        )

        su.evolve(33, tau=0.3)
        su.state = su.best["state"]
        su.evolve(33, tau=0.1)
        su.state = su.best["state"]
        su.evolve(33, tau=0.03)
        su.state = su.best["state"]

        assert su.best["energy"] < -6.25


class TestFullUpdate:
    @pytest.mark.parametrize("backend", ["numpy", pytorch_case])
    def test_heis_small(self, backend):
        Lx = 3
        Ly = 4
        D = 2

        psi0 = qtn.PEPS.rand(Lx, Ly, D)
        ham = qtn.ham_2d_heis(Lx, Ly)

        def to_backend(x):
            import autoray

            return autoray.do("array", x, like=backend)

        psi0.apply_to_arrays(to_backend)
        ham.apply_to_arrays(to_backend)

        su = qtn.FullUpdate(
            psi0, ham, progbar=True, keep_best=True, compute_energy_every=1
        )

        su.evolve(33, tau=0.3)
        su.state = su.best["state"]
        su.evolve(33, tau=0.1)
        su.state = su.best["state"]
        su.evolve(33, tau=0.03)
        su.state = su.best["state"]

        assert su.best["energy"] < -6.30
